package lib.java;


import java.util.*;

import errors.BadAttributeException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import types.WidgetDef.Derived;
import util.Builders;

/**
 * Button Definition for a Java button.
 */

// @EName Implements Java's clickable button.
@Derived(Clickable.class)
public class Button extends WidgetDef {
	
	public Button(Widget w, WidgetDef parent){
		super(w, parent);
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#getLanguage()
	 */
	public String getLanguage() {
		return "Java";
	}
	
	
	@Override
	public List<String> getVariables() {
		 return Collections.singletonList(
	                String.format("public JButton %s;", id));
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#preamble()
	 */
	public String preamble(){
		return id.concat(" = new JButton();\n");
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#postamble()
	 */
	public String postamble(){
		return "";
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#addAttributes(interpreter.GUIAttributes)
	 */
	public String addAttributes(GUIAttributes attrs) throws 
	BadAttributeException {
		final StringBuilder builder = new StringBuilder();
		// @EName label Sets the button's text.
        if (attrs.getString("label") != null)
            builder.append(String.format("%s.setText(\"%s\");\n",id,
                    attrs.getString("label")));
		// @EName height Sets the button's height.
		final int height = attrs.getInteger("height");
		// @EName width Sets the button's width.
		final int width = attrs.getInteger("width");
        if (height != -1 && width != -1)
            builder.append(String.format(
                    "%s.setPreferredSize(new Dimension(%d,%d));\n",
                    id, width, height));
		// @EName selected Changes whether the button is selected.
		builder.append(String.format("%s.setSelected(%s);\n", id, 
				attrs.getBoolean("selected")));
		// @EName borderPainted Changes whether the border is painted.
		builder.append(String.format("%s.setBorderPainted(%s);\n", id,
				attrs.getBoolean("borderPainted")));
		// @EName contentAreaFilled Changes whether the content area is filled.
		builder.append(String.format("%s.setContentAreaFilled(%s);\n", id,
				attrs.getBoolean("contentareafilled")));
		// @EName enabled Changes whether the button is enabled.
		builder.append(String.format("%s.setEnabled(%s);\n", id,
				attrs.getBoolean("enabled")));
		// @EName horizontalAlignment Adjusts the horizontal alignment
		builder.append(String.format("%s.setHorizontalAlignment(%d);\n", id,
				attrs.getInteger("horizontalalignment")));
		// @EName verticalAlignment Adjusts the vertical alignment
		builder.append(String.format("%s.setVerticalAlignment(%d);\n", id,
				attrs.getInteger("verticalalignment")));
		// @EName backgroundColor Adjusts the background color.
        if (attrs.getString("backgroundColor") != null)
            builder.append(String.format("%s.setBackground(%s);\n", id,
                    attrs.getColor("backgroundColor")));
        
		return builder.toString();
	}
	
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#getDefaults()
	 * 
	 * This is where the defaults are defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getDefaults() has to do is return this field. 
	 */
	
    private static final Map<String, List<String>> defaults =    
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .put("label", (String) null)      // @EDefaultValue label null
			.put("height", "-1")              // @EDefaultValue height -1
			.put("width", "-1")               // @EDefaultValue width -1
            .put("selected", "false")         // @EDefaultValue selected false
            .put("borderPainted", "true")     // @EDefaultValue borderPainted true
            .put("contentareafilled", "true") // @EDefaultValue contentAreaFilled true
            .put("enabled", "true")           // @EDefaultValue enabled true
            .put("verticalalignment", "0")    // @EDefaultValue verticalAlignment 0 (Corresponds to SwingConstants.CENTER)
            .put("horizontalalignment", "0")  // @EDefaultValue horizontalAlignment 0 (Corresponds to SwingConstants.CENTER)
            .put("backgroundColor", (String) null) // @EDefaultValue backgroundColor null
            .finish();
    
    public Map<String,List<String>> getDefaults() {
		return defaults;
	}
		
    private static final Map<String, String> synonyms =
        Builders.buildMap(new TreeMap<String, String>())
            .put("text", "label")
            .finish();
    
    @Override
    public Map<String, String> getSynonymousAttributes() {
        return synonyms;
    }
    
	 /* This is where the modifier code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getModifiers() has to do is return this field. */
	
	private static final ArrayList<String> modifiers = Builders.makeCollection(
			new ArrayList<String>(),
			"public String getButtonLabel(JButton button) {\n" +
            "\treturn button.getText();\n" +
            "}\n" +
            
            "public void setButtonLabel(JButton button, String newLabel) {\n" +
            "\tbutton.setText(newLabel);\n" +
            "}\n",
			
            "public String getButtonText(JButton button) {\n" +
            "\treturn getButtonLabel(button);\n" +
            "}\n" +
            
            "public void setButtonText(JButton button, String newText) {\n" +
            "\tsetButtonLabel(button, newText);\n" +
            "}\n",
            
            "public int getButtonHeight(JButton button) {\n" +
			"\treturn button.getHeight();\n" +
			"}\n" +
			
			"public void setButtonHeight(JButton button, int newHeight) {\n" +
			"\tDimension newDimension = new Dimension(button.getWidth(),newHeight);\n" +
			"\tbutton.setPreferredSize(newDimension);\n" +
			"}\n",
			
			"public int getButtonWidth(JButton button) {\n" +
			"\treturn button.getWidth();\n" +
			"}\n" +
			
			"public void setButtonWidth(JButton button, int newWidth) {\n" +
			"\tDimension newDimension = new Dimension(newWidth,button.getHeight());\n" +
			"\tbutton.setPreferredSize(newDimension);\n" +
			"}\n",
			
			"public Boolean getButtonSelected(JButton button) {\n" +
			"\treturn button.isSelected();\n" +
			"}\n" +
			
			"public void setButtonSelected(JButton button, boolean newSelected) {\n" +
			"\tbutton.setSelected(newSelected);\n" +
			"}\n",
			
			"public Boolean getButtonBorderPainted(JButton button) {\n" +
			"\treturn button.isBorderPainted();\n" +
			"}\n" +
			
			"public void setButtonBorderPainted(JButton button, boolean newBorderPainted) {\n" +
			"\tbutton.setBorderPainted(newBorderPainted);\n" +
			"}\n",
			
			"public Boolean getButtonContentAreaFilled(JButton button) {\n" +
			"\treturn button.isContentAreaFilled();\n" +
			"}\n" +
			
			"public void setButtonContentAreaFilled(JButton button, boolean newContentAreaFilled) {\n" +
			"\tbutton.setContentAreaFilled(newContentAreaFilled);\n" +
			"}\n",
			
			
			"public Boolean getButtonEnabled(JButton button) {\n" +
			"\treturn button.isEnabled();\n" +
			"}\n" +
			
			"public void setButtonEnabled(JButton button, boolean newEnabled) {\n" +
			"\tbutton.setEnabled(newEnabled);\n" +
			"}\n",
			
			"public int getButtonVerticalAlignment(JButton button) {\n" +
			"\treturn button.getVerticalAlignment();\n" +
			"}\n" +
			
			"public void setButtonVerticalAlignment(JButton button, int newAlignment) {\n" +
			"\tbutton.setVerticalAlignment(newAlignment);\n" +
			"}\n",
			
			"public int getButtonHorizontalAlignement(JButton button) {\n" +
			"\treturn button.getHorizontalAlignment();\n" +
			"}\n" +
			
			"public void setButtonHorizontalAlignment(JButton button, int newAlignment) {\n" +
			"\tbutton.setHorizontalAlignment(newAlignment);\n" +
			"}\n",
			
			"public Color getButtonBackgroundColor(JButton button) {\n" +
			"\treturn button.getBackground();\n" +
			"}\n" +
			
			"public void setButtonBackgroundColor(JButton button, Color newColor) {\n" +
			"\tbutton.setBackground(newColor);\n" +
			"}\n"
			
	);
	
	public ArrayList<String> getModifiers() {
		return modifiers;
	}
		
	/* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
	 private static final Set<String> imports = Builders.makeCollection(
	            new TreeSet<String>(), 
	            "import javax.swing.JButton;\n",
	            "import java.awt.Dimension;\n",
	            "import java.awt.Color;\n"
	    );
	
	public Set<String> getImports() {
		return imports;
	}
}