package lib.java;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;


import errors.BadAttributeException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import util.Builders;

/**
 * @EName A Non-Swing clock widget rendered on a JPanel.
 */ 
public class Clock extends WidgetDef {

	String backgroundColor;
	String textColor;
	int fontSize;
	boolean seconds;
	
	public Clock(Widget widget, WidgetDef parent) {
		super(widget, parent);
	}
	
	public List<String> getVariables(){
		return Builders.makeCollection(new ArrayList<String>(),
					String.format("public GUIClock %s;\n",id));
	}
	
	@Override
	public String getLanguage() {
		return "Java";
	}

	@Override
	public String preamble() {
		return String.format("%s = new GUIClock(%s,%s,%d,%b);\n",widget.id,
					backgroundColor,textColor,fontSize,seconds);
	}
	/*
	 *  (non-Javadoc)
	 * @see types.WidgetDef#getDefaults()
	 * 
	 * This is where the defaults are defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getDefaults() has to do is return this field. 
	 */
	
    private static final Map<String, List<String>> defaults =    
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .put("backgroundcolor", "white")	//@EDefaultValue backgroundcolor white
            .put("textcolor","black") 			//@EDefaultValue textcolor black
            .put("size","18")					//@EDefaultValue size 18
            .put("seconds","false")				//@EDefaultValue seconds false
            .finish();
	@Override
	public Map<String, List<String>> getDefaults() {
		return defaults;
	}

	@Override
	public String addAttributes(GUIAttributes attributes)
			throws BadAttributeException {
		final StringBuilder builder = new StringBuilder();
		//@EName backgroundcolor Sets the background color of the clock.
		backgroundColor = attributes.getColor("backgroundcolor"); 
		//@EName textcolor Sets the text color of the clock.
		textColor = attributes.getColor("textcolor");
		//@EName size Specifies the font size of the clock text.
		fontSize = attributes.getInteger("size");
		//@EName seconds Specifies whether or not to include the seconds on the clock.
		seconds = attributes.getBoolean("seconds");
		
		return builder.toString();
	}

	@SuppressWarnings("unchecked")
	@Override
	public List<String> getModifiers() {
		return Collections.EMPTY_LIST;
	}

	/* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
	 private static final Set<String> imports = Builders.makeCollection(
	            new TreeSet<String>(), 
	            "java.awt.Color",
	            "java.awt.Dimension",
	            "java.awt.Font",
	            "java.awt.FontMetrics",
	            "java.awt.Graphics",
	            "java.awt.font.FontRenderContext",
	            "java.awt.geom.Rectangle2D",
	            "java.text.SimpleDateFormat",
	            "java.util.Date",
	            "java.util.Timer",
	            "java.util.TimerTask",
	            "javax.swing.JPanel"
	    );
	
	public Set<String> getImports() {
		return imports;
	}

	public Set<String> getInnerClasses(){
		final StringBuilder builder = new StringBuilder();		
		
		builder.append("public class GUIClock extends JPanel {\n");
		builder.append("\tString time;\n");
		builder.append("\tString format;\n");
		builder.append("\tTimer timer;\n");
		builder.append("\tFont clockFont;\n");
		builder.append("\tColor backgroundColor;\n");
		builder.append("\tColor textColor;\n");
		builder.append("\tlong refreshRate = 1000;\n");
		builder.append("\tpublic GUIClock(Color bC, Color tC, int size, boolean seconds){\n");
		builder.append("\t\tif (seconds) { format=\"hh:mm:ss\"; time = \"00:00:00\"; }\n");
		builder.append("\t\telse { format = \"hh:mm\";  time = \"00:00\"; }\n");
		builder.append("\t\tbackgroundColor = bC;\n");
		builder.append("\t\ttextColor = tC;\n");
		builder.append("\t\tclockFont = new Font(\"Footlight MT Light\", Font.BOLD, size);\n");
		builder.append("\t\ttimer = new Timer();\n");
		builder.append("\t\ttimer.scheduleAtFixedRate(new ClockTick(this),0,refreshRate);\n");
		builder.append("\t\tRectangle2D bounds = clockFont.getStringBounds(time,new FontRenderContext(null,true,true));\n");
		builder.append("\t\tsetPreferredSize(new Dimension((int)bounds.getWidth()+10,(int)bounds.getHeight()+10));\n");
		builder.append("\t}\n");
		
		builder.append("\tpublic void setRefreshRate(long refreshRate){\n");
		builder.append("\t\ttimer.cancel();");
		builder.append("\t\ttimer = new Timer();\n");
		builder.append("\t\ttimer.scheduleAtFixedRate(new ClockTick(this),0,refreshRate);\n");
		builder.append("\t}");
		
		builder.append("\tpublic void killClock(){\n");
		builder.append("\t\ttimer.cancel();\n");
		builder.append("\t}\n");
		
		builder.append("\tprotected void paintComponent(Graphics g){\n");
		builder.append("\t\tg.setFont(clockFont);\n");
		builder.append("\t\tFontMetrics metrics = g.getFontMetrics();\n");
		builder.append("\t\tRectangle2D bounds = g.getFontMetrics().getStringBounds(time,g);\n");
		builder.append("\t\tg.setColor(backgroundColor);\n");
		builder.append("\t\tg.fillRect(-5,-5,(int)bounds.getWidth()+20,(int)bounds.getHeight()+20);\n");
		builder.append("\t\tg.setColor(textColor);\n");
		builder.append("\t\tg.drawString(time,5,(metrics.getHeight()+ metrics.getAscent()+8)/2);\n");
		builder.append("\t}\n");
		
		builder.append("\tpublic class ClockTick extends TimerTask{\n");
		builder.append("\t\tGUIClock clock;\n");
		builder.append("\t\tpublic ClockTick(GUIClock clock){\n");
		builder.append("\t\t\tthis.clock = clock;\n");
		builder.append("\t\t}\n");
		builder.append("\t\tpublic void run() {\n");
		builder.append("\t\t\tDate date = new Date();\n");
		builder.append("\t\t\tSimpleDateFormat formatter = new SimpleDateFormat(clock.format);\n");
		builder.append("\t\t\tclock.time = formatter.format(date);\n");
		builder.append("\t\t\tclock.repaint();\n");
		builder.append("\t\t}\n");
		builder.append("\t}\n}");

		return Builders.makeCollection(new TreeSet<String>(), builder.toString());
	}

}
