package lib.java;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import errors.BadAttributeException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import util.Builders;

/**
 * @EName Implements Java's JTextArea class.
 *
 */
public class TextArea extends WidgetDef {
	boolean lineNum = false;
	String textAreaId;
	
	public TextArea(Widget widget, WidgetDef parent) {
        super(widget, parent);
        
		// linenumbers is a special attribute, since it affects the whole
		// widget description.  It is processed here, and *not* in the
		// addAttributes method.
        //
        //@EName linenumbers Boolean that sets whether the linenumbers can be seen (true)
        // or not (false).
        //
		if(widget.attrs.get("linenumbers")!=null)
			lineNum = true;
		
		if(lineNum)
			textAreaId = id + ".accessibleTextComponent";
		else
			textAreaId = id;
		
	}

	@Override
	public String getLanguage() {
		return "Java";
	}
	
	@Override
	public List<String> getVariables() {
		if(!lineNum)
			return Collections.singletonList(
	                String.format("public JTextArea %s;\n", id));
		else
			return Collections.singletonList(
						String.format("public SmartTextArea %s;\n", id));
	}

	@Override
	public String preamble() {
		if(!lineNum)
			return id.concat(" = new JTextArea();\n");
		else
			return id.concat(" = new SmartTextArea();\n");
	}

	@Override
	public String postamble() {
		return "";
	}

	@Override
	public String addAttributes(GUIAttributes attributes) throws BadAttributeException {
		final StringBuilder builder = new StringBuilder();
		
		//@EName textHeight Sets the height of the TextArea (in character sizes, not pixels)
		if (attributes.getInteger("textHeight") != -1)
			builder.append(String.format("%s.setRows(%d);\n",textAreaId,
					attributes.getInteger("textHeight")));
		//@EName textWidth Sets the width of the TextArea (in character sizes, not pixels)
		if (attributes.getInteger("textWidth") != -1)
			builder.append(String.format("%s.setColumns(%d);\n",textAreaId,
					attributes.getInteger("textWidth")));
		//@EName text Sets the initial text in the TextArea
		if (attributes.getOne("text") != null)
			builder.append(String.format("%s.setText(\"%s\");\n", textAreaId,
					attributes.getOne("text")));
		//@EName linewrap Boolean that sets whether the line wrap is enabled (true) or disabled (false).
		builder.append(String.format("%s.setLineWrap(%b);\n", textAreaId,
				attributes.getBoolean("linewrap")));
		if (attributes.getBoolean("linewrap"))
			builder.append(String.format("%s.setWrapStyleWord(true);\n", textAreaId));
		//@EName editable Boolean that sets whether the TextArea is editable (true) or not editable (false).
		builder.append(String.format("%s.setEditable(%b);\n", textAreaId,
				attributes.getBoolean("editable")));
		//@EName tabSize Sets the tab size in character widths.
		if (attributes.getInteger("tabSize")!= -1)
			builder.append(String.format("%s.setTabSize(%d);\n", textAreaId, 
				attributes.getInteger("tabSize")));
		
		return builder.toString();
	}

	/*
	 * This is where the defaults are defined. The idea is to have a
	 * private static field that gets initialized at class loading time with
	 * all the code; this way it doesn't get rebuilt each time. Then all that
	 * getDefaults() has to do is return this field. 
	 */

	private static final Map<String, List<String>> defaults =    
	    Builders.buildMultiMap(new TreeMap<String, List<String>>())
	    .put("textHeight","-1") 		//@EDefaultValue height -1
	    .put("textWidth","-1") 			//@EDefaultValue textWidth -1
	    .put("text",(String) null) 	//@EDefaultValue text null
	    .put("linewrap","true")		//@EDefaultValue linewrap true
	    .put("editable","true")		//@EDefaultValue editable true
	    .put("tabSize","-1")		//@EDefaultValue tabSize -1
	    .put("linenumbers","false")	//@EdefaultValue linenumber false
	    .finish();
	
	@Override
	public Map<String, List<String>> getDefaults() {
		return defaults; 
	}

	 /* This is where the modifier code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getModifiers() has to do is return this field. */
	private static final ArrayList<String> modifiers = Builders.makeCollection(
			new ArrayList<String>(),
			"public int getTextAreaHeight(JTextArea area) {\n" +
			"\treturn area.getRows();\n" +
			"}\n" +
			
			"public void setTextAreaHeight(JTextArea area, int newHeight) {\n" +
			"\tarea.setRows(newHeight);\n" +
			"}\n",
			
			"public int getTextAreaWidth(JTextArea area) {\n" +
			"\treturn area.getColumns();\n" +
			"}\n" +
			
			"public void setTextAreaWidth(JTextArea area, int newWidth) {\n" +
			"\tarea.setColumns(newWidth);\n" +
			"}\n",
			
			"public String getTextAreaText(JTextArea area) {\n" +
			"\treturn area.getText();\n" +
			"}\n" +
			
			"public void setTextAreaText(JTextArea area, String newText) {\n" +
			"\tarea.setText(newText);\n" +
			"}\n",
			
			"public boolean getTextAreaLineWrap(JTextArea area) {\n" +
			"\treturn area.getLineWrap();\n" +
			"}\n" +
			
			"public void setTextAreaLineWrap(JTextArea area, boolean newLineWrap) {\n" +
			"\tarea.setLineWrap(newLineWrap);\n" +
			"}\n",
			
			"public boolean getTextAreaEditable(JTextArea area) {\n" +
			"\treturn area.isEditable();\n" +
			"}\n" +
			
			"public void setTextAreaEditable(JTextArea area, boolean newEditable) {\n" +
			"\tarea.setEditable(newEditable);\n" +
			"}\n"
			);
	
	private static final ArrayList<String> lineModifiers = Builders.makeCollection(
			new ArrayList<String>(),
			"public int getTextAreaHeight(SmartTextArea area) {\n" +
			"\treturn area.accessibleTextComponent.getRows();\n" +
			"}\n" +
			
			"public void setTextAreaHeight(SmartTextArea area, int newHeight) {\n" +
			"\tarea.accessibleTextComponent.setRows(newHeight);\n" +
			"}\n",
			
			"public int getTextAreaWidth(SmartTextArea area) {\n" +
			"\treturn area.accessibleTextComponent.getColumns();\n" +
			"}\n" +
			
			"public void setTextAreaWidth(SmartTextArea area, int newWidth) {\n" +
			"\tarea.accessibleTextComponent.setColumns(newWidth);\n" +
			"}\n",
			
			"public String getTextAreaText(SmartTextArea area) {\n" +
			"\treturn area.accessibleTextComponent.getText();\n" +
			"}\n" +
			
			"public void setTextAreaText(SmartTextArea area, String newText) {\n" +
			"\tarea.accessibleTextComponent.setText(newText);\n" +
			"}\n",
			
			"public boolean getTextAreaLineWrap(SmartTextArea area) {\n" +
			"\treturn area.accessibleTextComponent.getLineWrap();\n" +
			"}\n" +
			
			"public void setTextAreaLineWrap(SmartTextArea area, boolean newLineWrap) {\n" +
			"\tarea.accessibleTextComponent.setLineWrap(newLineWrap);\n" +
			"}\n",
			
			"public boolean getTextAreaEditable(SmartTextArea area) {\n" +
			"\treturn area.accessibleTextComponent.isEditable();\n" +
			"}\n" +
			
			"public void setTextAreaEditable(SmartTextArea area, boolean newEditable) {\n" +
			"\tarea.accessibleTextComponent.setEditable(newEditable);\n" +
			"}\n"
			);
	
	@Override
	public ArrayList<String> getModifiers() {
		if(!lineNum)
			return modifiers;
		else
			return lineModifiers;
	}

	/* This is where the import code is defined. The idea is to have a
     * private static field that gets initialized at class loading time with
     * all the code; this way it doesn't get rebuilt each time. Then all that
     * getImports() has to do is return this field. */
	private static final Set<String> imports = Builders.makeCollection(
	        new TreeSet<String>(),
	        "javax.swing.JTextArea",
	        "javax.swing.event.*",
	        "java.awt.event.*",
	        "java.awt.Point"
	);
	 
    private static final Set<String> lineNumImports = Builders.makeCollection(
            new TreeSet<String>(imports),
            "javax.swing.JPanel",
            "javax.swing.JScrollPane",
            "java.awt.FontMetrics",
            "java.awt.GridLayout",
            "java.awt.BorderLayout",
            "java.awt.Graphics",
            "javax.swing.SwingUtilities"
            );
     
	@Override
	public Set<String> getImports() {
		return lineNum ? lineNumImports : imports;
	}
	
	public Set<String> getInnerClasses() {
		if(lineNum){
			StringBuilder builder = new StringBuilder();
			builder.append("/*\n");
			builder.append(" * Name: SmartTextArea\n");
			builder.append(" * Description:A utility tool I can call this JComponent.\n");
			builder.append(" * I hated JTextArea that it does not have internal capability of displaying line numbers, and one click line selection.\n");
			builder.append(" * I needed when I was developing an ide for MIPS language as source code editor.\n");
			builder.append(" * Use it like your own :)\n");
			builder.append(" * By: Fatih Keles\n");
			builder.append(" * \n");
			builder.append("*\n"); 
			builder.append(" * Inputs:several constructers are available but empty constructer works fine too.\n");
			builder.append(" * Also has a test main method inside code\n");
			builder.append("*\n"); 
			builder.append(" *  Returns:A JPanel that can be plugged anywhere.\n");
			builder.append(" *\n"); 
			builder.append(" * Assumes:One constructer also takes a java.awt.Point in it that is updated as x has the line number, y has the char number in text area.\n");
			builder.append(" * may be usefull to follw caret positions as I used for that purpose\n");
			builder.append("*\n"); 
			builder.append("* Side Effects:not known\n");
			builder.append(" * This code is copyrighted and has limited warranties.\n");
			builder.append("* Please see http://www.Planet-Source-Code.com/xq/ASP/txtCodeId.5095/lngWId.2/qx/vb/scripts/ShowCode.htm\n");
			builder.append("* for details.\n");
			builder.append("*/\n");
	
			builder.append("/**\n");
			builder.append("*SmartArea.java is a somehow skilled java text component.\n");
			builder.append("*Pluggable into any swing application.\n");
			builder.append("*I did not made the awt compatibility tests yet.\n");
			builder.append("*if anyone improves or makes changes over this\n"); 
			builder.append("*please send me a copy to: fatihkeles@gmail.com\n");
			builder.append("*It can show the text with line numbers, and select entire\n"); 
			builder.append("*row by clicking number pane on left;\n");
			builder.append("*/\n");
	
			builder.append("/*\n");
			builder.append("*@author fkeles\n");
			builder.append("*@version 0.1 Key listener for numbering & caret listener used for display and selection\n");
			builder.append("*@version 0.2 key listener replaced with document listener\n");
			builder.append("*\n");
			builder.append("*/\n");
			builder.append("public class SmartTextArea extends JPanel{\n");
			builder.append("    	private JTextArea accessibleTextComponent;\n");
			builder.append("		private NumberPane numberPane;\n");
			builder.append("    	private JScrollPane scroller;\n");
			builder.append("    	private JPanel rootPane,contentPane;\n");
			builder.append("    	private Point caretPosition;\n");
	
			builder.append("       	public SmartTextArea(){\n");
			builder.append("       		contentPane=new JPanel(new BorderLayout());\n");
			
			builder.append("       		scroller=new JScrollPane(contentPane);\n");
			builder.append("       		accessibleTextComponent=new JTextArea();\n");
			builder.append("       		numberPane=new NumberPane();\n");
			builder.append("       		rootPane=this;\n");
			builder.append("       		caretPosition=new Point();\n");
			builder.append("			super.setLayout(new GridLayout(1,1));\n");
			       		
			builder.append("       		contentPane.add(accessibleTextComponent,BorderLayout.CENTER);\n");
			builder.append("       		contentPane.add(numberPane,BorderLayout.WEST);\n");
			       		
			builder.append("       		rootPane.add(scroller);\n");
			builder.append("       	}\n");
	
	
			builder.append("        public SmartTextArea(Point caretReports){\n");
			builder.append("           	this();\n");
			builder.append("           	caretPosition=caretReports;\n");
			builder.append("        }\n");
	
	
			builder.append("       	public SmartTextArea(String text){\n");
			builder.append("       		this();\n");
			builder.append("       		setText(text);\n");
			builder.append("       	}\n");
			                	
			builder.append("       	public JTextArea getAccessibleTextComponent(){\n");
			builder.append("       		return accessibleTextComponent;\n");
			builder.append("      	}\n");
	
			builder.append("         public String getText(){\n");
			builder.append("           	return accessibleTextComponent.getText();\n");
			builder.append("         }\n");
	
	
			builder.append("         public void setText(String text){\n");
			builder.append("         	accessibleTextComponent.setText(text);\n");
			builder.append("         }\n");
			
			builder.append("         public int getRowCount(){\n");
			builder.append("         	return accessibleTextComponent.getRows();\n");
			builder.append("         }\n");
	
			builder.append("		public void append(String text){\n");
			builder.append("			accessibleTextComponent.append(text);\n");
			builder.append("		}\n");
			
			builder.append("		public void copy(){\n");
			builder.append("			accessibleTextComponent.copy();\n");
			builder.append("		}\n");
			
			builder.append("		public String getSelectedText(){\n");
			builder.append("			return accessibleTextComponent.getSelectedText();\n");
			builder.append("		}\n");
			
			builder.append("		public void cut(){\n");
			builder.append("			accessibleTextComponent.cut();\n");
			builder.append("		}\n");
			
			builder.append("		public void paste(){\n");
			builder.append("			accessibleTextComponent.paste();\n");
			builder.append("		}\n");
			
			builder.append("		public void selectAll(){\n");
			builder.append("			accessibleTextComponent.selectAll();\n");
			builder.append("		}\n");
	
			builder.append("         public class NumberPane extends JPanel implements DocumentListener,MouseListener,CaretListener{\n");
			builder.append("         	private int width=20,lineCount=0,d_height=0;\n");
			builder.append("            private Color foregroundColor;\n");
			builder.append("            private Color defaultForegroundColor=new Color(128,0,0);\n");
			builder.append("            private int selection=-1;\n");
	
	
			builder.append("            public NumberPane(){\n");
			builder.append("               foregroundColor=defaultForegroundColor;\n");
			builder.append("               setPreferredSize(new Dimension(width,-1));\n");
			builder.append("               addMouseListener(this);\n");
			builder.append("               accessibleTextComponent.getDocument().addDocumentListener(this);\n");
			builder.append("               accessibleTextComponent.addCaretListener(this);\n");
			builder.append("            }\n");
			            
			builder.append("            public void insertUpdate(DocumentEvent de){repaint();}\n");
			builder.append("            public void removeUpdate(DocumentEvent de){repaint();}\n");
			builder.append("            public void changedUpdate(DocumentEvent de){repaint();}\n");
			                                        		
			builder.append("            public void mouseClicked(MouseEvent me){}\n");
			builder.append("            public void mouseReleased(MouseEvent me){}\n");
			builder.append("            public void mouseEntered(MouseEvent me){}\n");
			builder.append("            public void mouseExited(MouseEvent me){}\n");
	
	
			builder.append("            public void mousePressed(MouseEvent me){\n");
			builder.append("            	if(SwingUtilities.isLeftMouseButton(me)){//only left click\n");
			builder.append("            		selection=(me.getY())/d_height;\n");
			builder.append("            		if(selection>=0&&selection<lineCount){\n");
			builder.append("            			select(selection);\n");
			builder.append("            			selection=-1;\n");
			builder.append("            		}\n");
			builder.append("            	}\n");
			builder.append("            }\n");
			
			builder.append("         	public void caretUpdate(CaretEvent ce){\n");
			builder.append("          		try{\n");
			builder.append("          			int l=accessibleTextComponent.getLineOfOffset(ce.getDot());\n");
			builder.append("          			int c=accessibleTextComponent.getLineStartOffset(l);\n");
			builder.append("           			l+=1;\n");
			builder.append("           			c=ce.getDot()-c;\n");
			
			builder.append("           			caretPosition.x=l;\n");
			builder.append("           			caretPosition.y=c;\n");
			builder.append("          		}catch(Exception e){}\n");
			builder.append("         	}\n");
			                                                        		
			builder.append("         	public void select(int i){\n");
			builder.append("          		try{\n");
			builder.append("           			int s1=accessibleTextComponent.getLineStartOffset(i);\n");
			builder.append("           			int s2=accessibleTextComponent.getLineEndOffset(i);\n");
			builder.append("           			accessibleTextComponent.setCaretPosition(s1);\n");
			builder.append("           			accessibleTextComponent.moveCaretPosition(s2);\n");
	
			builder.append("          		}catch(Exception r){\n");
			builder.append("                	System.out.println(r);\n");
			builder.append("            	}\n");
			builder.append("         	}\n");
			                                                                    				
			builder.append("         	public void paint(Graphics g){\n");
			builder.append("         		super.paint(g);\n");
			builder.append("            	Color old=g.getColor();\n");
			builder.append("            	g.setColor(foregroundColor);\n");
			builder.append("            	lineCount=accessibleTextComponent.getLineCount();\n");
			builder.append("            	FontMetrics fm=accessibleTextComponent.getFontMetrics(accessibleTextComponent.getFont());\n");
			builder.append("            	d_height=fm.getHeight();\n");
			builder.append("            	int pwidth=fm.stringWidth(\"\"+(lineCount+1))+4;\n");
	
			builder.append("            	if(pwidth-width>=0){\n");
			builder.append("            		width+=10;\n");
			builder.append("            		setPreferredSize(new Dimension(width,-1));\n");
			builder.append("            	}else{\n");
			builder.append("            		if(pwidth-width<-10){\n");
			builder.append("            			width-=10;\n");
			builder.append("            			setPreferredSize(new Dimension(width,-1));\n");
			builder.append("            		}\n");
			builder.append("            	}\n");
			            
			builder.append("            	for(int i=0;i<lineCount+1;i++){\n");
			builder.append("            		g.drawString((i+1)+\"\",2,i*d_height+d_height-4);\n");
			builder.append("            	}\n");
			builder.append("            	g.setColor(old);\n");
			builder.append("         	}\n");
			builder.append("         }\n");
			builder.append("}\n");
			
			return Builders.buildCollection(new TreeSet<String>())
						.add(builder.toString()).finish();
		} else
			return Collections.emptySet();
	}

}
