package lib.java;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import errors.BadAttributeException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import util.Builders;

/**
 * @EName Widget for an editable Java JTextField.
 *
 */
public class TextField extends WidgetDef {

	public TextField(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }
    
	@Override
	public String getLanguage() {
		return "Java";
	}
	
	@Override
	public List<String> getVariables() {
		 return Collections.singletonList(
	                String.format("public JTextField %s;\n", id));
	}

	@Override
	public String preamble() {
		return id.concat(" = new JTextField();\n");
	}

	@Override
	public String postamble() {
		return "";
	}

	@Override
	public String addAttributes(GUIAttributes attributes) throws BadAttributeException {
		final StringBuilder builder = new StringBuilder();
		//@EName textWidth sets the width of the TextField (with character, not pixel, width)
		if (attributes.getInteger("textWidth") != -1)
			builder.append(String.format("%s.setColumns(%d);\n",id,
					attributes.getInteger("textWidth")));
		//@EName alignHoriz Sets the horizontal alignment
		builder.append(String.format(
                "%s.setHorizontalAlignment(SwingConstants.%s);\n", id,
				attributes.getEnum("alignHoriz", Alignment.class).constId));
        //@EName editable Boolean that sets whether the TextField is editable (true) or not editable (false).
        builder.append(String.format("%s.setEditable(%s);\n", id,
                attributes.getBoolean("editable")));
		//@EName text Sets the initial text in the TextField.
		if (attributes.getOne("text") != null)
			builder.append(String.format("%s.setText(\"%s\");\n", id,
					attributes.getOne("text")));
		return builder.toString();
	}

    private enum Alignment {
        left("LEADING"), center("CENTER"), right("TRAILING");
        
        private final String constId;
        Alignment(String constId) {
            this.constId = constId;
        }
    }
    
	private static final Map<String, List<String>> defaults =    
	    Builders.buildMultiMap(new TreeMap<String, List<String>>())
	    .put("textWidth", "-1") 		//@EDefaultValue textWidth -1
	    .put("alignHoriz", "left")	//@EDefaultValue alignHoriz left
	    .put("editable", "true")	//@EDefaultValue editable true
        .put("text", (String) null)	//@EDefaultValue text null
	    .finish();
	
	@Override
	public Map<String, List<String>> getDefaults() {
	    return defaults; 
	}

	private static final ArrayList<String> modifiers = Builders.makeCollection(
			new ArrayList<String>(),
			"public int getTextFieldWidth(JTextField field) {\n" +
			"\treturn field.getColumns();\n" +
			"}\n" +
			
			"public void setTextFieldWidth(JTextField field, int newWidth) {\n" +
			"\tfield.setColumns(newWidth);\n" +
			"}\n",
			
			"public int getTextFieldHorizontalAlignment(JTextField field) {\n" +
			"\treturn field.getHorizontalAlignment();\n" +
			"}\n" +
			
			"public void setTextFieldHorizontalAlignment(JTextField field, int newAlignment) {\n" +
			"\tfield.setHorizontalAlignment(newAlignment);\n" +
			"}\n",
			
			"public String getTextFieldText(JTextField field) {\n" +
			"\treturn field.getText();\n" +
			"}\n" +
			
			"public void setTextFieldText(JTextField field, String newText) {\n" +
			"\tfield.setText(newText);\n" +
			"}\n",
            
            "public boolean getTextFieldEditable(JTextField field) {\n" +
            "\treturn field.isEditable();\n" +
            "}\n" +
            
            "public void setTextFieldEditable(JTextField field, boolean newValue) {\n" +
            "\tfield.setEditable(newValue);\n" +
            "}\n"
		);

	@Override
	public ArrayList<String> getModifiers() {
		return modifiers;
	}

	 private static final Set<String> imports =
         Builders.makeCollection(new TreeSet<String>(),
                 "javax.swing.JTextField",
                 "javax.swing.SwingConstants");

	@Override
	public Set<String> getImports() {
		return imports;
	}

}
