package lib.java;

import java.util.*;

import output.java.CodeBuilder;
import output.java.State;

import errors.BadAttributeException;
import errors.BadChildWidgetException;
import types.GUIAttributes;
import types.Widget;
import types.WidgetDef;
import util.Builders;

/**
 * @EName Widget that holds TreeNodes.
 * 
 * 
 * @author ritza
 *
 */
public class Tree extends WidgetDef {
	
	public Tree(Widget widget, WidgetDef parent) {
        super(widget, parent);
    }
    
	@Override
	public String getLanguage() {
		return "Java";
	}

	@Override
	public String preamble() {
		CodeBuilder str = new CodeBuilder();
		str.append("AUX$%s$root = new DefaultMutableTreeNode();\n", id);
		str.append("%s = new JTree(AUX$%1$s$root);\n", id);
		return str.toString();
	}
	
	public String postamble() {
		StringBuilder str = new StringBuilder();
		str.append(id.concat(".expandRow(0);\n"));
		str.append(id.concat(".validate();\n"));
		return str.toString();
	}

	@Override
	public String addAttributes(GUIAttributes attributes) throws BadAttributeException {
		final CodeBuilder builder = new CodeBuilder(State.METHOD);

		/*
		 * @EName text sets the text for the root node.
		 */
		if (attributes.getOne("label") != null)
			builder.append("AUX$%s$root.setUserObject(\"%s\");",
                    id, attributes.getString("label"));
		
		/*
		 * @EName onSelected If onSelected is specified, a TreeSelectionListener is
		 * implemented using the onSelected value as the method name.
		 */
		if (attributes.getOne("onSelected") == null)
			return builder.toString();
		
		String onSelected = attributes.getOne("onSelected");
		final String handlerClass =
			String.format("temp$%s$onSelectedHandler", id); 
		
		builder.beginBlock("")
		.append("final Component temp$%s = %1$s;", id)
		.append("final %s temp$controller = controller;",
				widget.file.getControllerId())
				.beginClass("class %s implements TreeSelectionListener", handlerClass)
				.beginMethod("public void valueChanged(TreeSelectionEvent event)")
				.append("if (temp$controller != null) temp$controller.%s(temp$%s, event);",
						onSelected, id)
						.finishMethod()
						.finishClass()
						.append("%s.addTreeSelectionListener(new %s());", id, handlerClass)
						.finishBlock();
		return builder.toString();
}

	private static final Map<String, List<String>> defaults =
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .put("onSelected", (String) null) 	//@EDefaultValue onSelected null
            .put("label",(String)null)			//@EDefaultValue label null
            .finish();
	
	@Override
	public Map<String, List<String>> getDefaults() {
		return defaults;
	}
	
    @Override
	public String addChild(String childType, String childName, 
            GUIAttributes childAttrs)
    throws BadChildWidgetException {
		if (isDerivedFrom(childType, TreeNode.class))
			return String.format("AUX$%s$root.add(%s);\n", id, childName);
		else
			throw new BadChildWidgetException(widget);
    }
    
  	@Override
	public List<String> getModifiers() {
		return Collections.emptyList();
	}
	
    private static final Map<String, String> synonyms =
        Builders.buildMap(new TreeMap<String, String>())
            .put("text", "label")
            .finish();
    
    @Override
    public Map<String, String> getSynonymousAttributes() {
        return synonyms;
    }

	private static final Set<String> imports = Builders.makeCollection(
            new TreeSet<String>(), 
            "import javax.swing.tree.DefaultMutableTreeNode;\n",
            "import javax.swing.event.TreeSelectionListener;\n",
            "import javax.swing.event.TreeSelectionEvent;\n",
            "import javax.swing.JTree;\n"
    );

	@Override
	public Set<String> getImports() {
		return imports;
	}

    @Override
    public List<String> getVariables() {
        return Builders.makeCollection(new ArrayList<String>(2),
                "public JTree " + id + ";\n",
                "public DefaultMutableTreeNode AUX$" + id + "$root;\n");
    }
}