package output.java;

import java.util.*;

import util.Functional;
import util.Pair;

final class ImportStatement implements Comparable<ImportStatement> {
    final List<String> parts;
    final boolean isStatic;
    final boolean isWildcard;
    
    ImportStatement(List<String> parts, boolean isStatic, boolean isWildcard) {
        this.parts = parts;
        this.isStatic = isStatic;
        this.isWildcard = isWildcard;
    }
    
    ImportStatement(String target, boolean isStatic) {
        this(parseTarget(target), isStatic, target.endsWith(".*"));
    }
    
    private static List<String> parseTarget(String target) {
        if (target.endsWith(".*"))
            target = target.substring(0, target.length() - 2);            
        return Arrays.asList(target.split("\\."));
    }

    public int compareTo(ImportStatement other) {
        for (Pair<String, String> ours_theirs : 
                Functional.zip(this.parts, other.parts)) {
            if (ours_theirs.a.equals(ours_theirs.b))
                continue;
            return ours_theirs.a.compareTo(ours_theirs.b);
        }
        
        if (this.parts.size() != other.parts.size())
            return this.parts.size() < other.parts.size() ? -1 : 1;
        
        if (this.isWildcard != other.isWildcard)
            return this.isWildcard ? 1 : -1;
        
        if (this.isStatic != other.isStatic)
            return this.isStatic ? 1 : -1;
        
        return 0;
    }
    
    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof ImportStatement))
            return super.equals(obj);
        final ImportStatement other = (ImportStatement) obj;
        
        if (this.parts.size() != other.parts.size() ||
                this.isStatic != other.isStatic ||
                this.isWildcard != other.isWildcard)
            return false;
        
        for (Pair<String, String> ours_theirs :
                Functional.zip(this.parts, other.parts))
            if (!ours_theirs.a.equals(ours_theirs.b))
                return false;
        
        return true;
    }
    
    boolean matches(ImportStatement other) {
        if (this.isStatic != other.isStatic)
            return false;
        
        if (!this.isWildcard)
            if (other.isWildcard)
                return other.matches(this);
            else 
                return this.equals(other);
        
        if (other.isWildcard)
            return this.equals(other);
        
        // Now we know that this is a wildcard and the other is not.
        
        if (this.parts.size() != other.parts.size() - 1)
            return false;
        
        for (Pair<String, String> ours_theirs :
            Functional.zip(this.parts, other.parts))
        if (!ours_theirs.a.equals(ours_theirs.b))
            return false;
    
        return true;
    }
    
    ImportStatement successor() {
        final List<String> newParts = new ArrayList<String>(
                parts.subList(0, parts.size() - 1));
        newParts.add(parts.get(parts.size() - 1) + 'a');
        return new ImportStatement(newParts, isStatic, isWildcard);
    }
    
    ImportStatement matchingWildcard() {
        if (isWildcard)
            return null;
        
        return new ImportStatement(parts.subList(0, parts.size() - 1),
                isStatic, true);
    }
    
    public String toString() {
        return String.format("import %s%s%s;", 
                isStatic ? "static " : "",
                Functional.join(parts, "."),
                isWildcard ? ".*" : "");
    }
}
