package output.java;

/**
 * The state of a {@link CodeBuilder} object. What state the builder is in
 * constrains the scopes that can be begun; for instance, a method cannot be
 * defined immediately within another.
 * <p>
 * The valid state transitions can be summarized thus:
 * <ul>
 *   <li>{@link #TOPLEVEL} &rarr; {@link #CLASS}</li>
 *   <li>{@link #CLASS} &rarr; {@link #METHOD} or {@link #CLASS}</li>
 *   <li>{@link #METHOD} &rarr; {@link #BLOCK} or {@link #CLASS}</li>
 *   <li>{@link #BLOCK} &rarr; {@link #BLOCK} or {@link #CLASS}</li>
 * </ul>
 * Also, any state can transition to {@link #COMMENT}, but {@link #COMMENT}
 * cannot transition to any other state. (This reflects the fact that a comment
 * can appear anywhere but cannot contain anything but plain text.)
 * <p>
 * When a scope is closed using a method such as 
 * {@link CodeBuilder#finishClass()}, the builder reverts to its previous
 * state, following one of the above transitions in reverse.
 */
public enum State {
    /**
     * The top level of the file, containing the package statement, import
     * statements, and classes.
     * <p>
     * Cannot be contained; can contain only {@link #CLASS} and 
     * {@link #COMMENT} scopes.
     */
    TOPLEVEL {
        boolean canContain(State state) {
            switch (state) {
                case CLASS:
                case COMMENT:
                    return true;
                default:
                    return false;
            }
        }
    },
    /**
     * Any class declaration (top-level, inner, local, etc.).
     * <p>
     * Can be contained by any scope; can contain {@link #CLASS},
     * {@link #METHOD}, and {@link #COMMENT} scopes.
     */
    CLASS {
        boolean canContain(State state) {
            switch (state) {
                case CLASS:
                case METHOD:
                case COMMENT:
                    return true;
                default:
                    return false;
            }
        }
    },
    /**
     * The body of a method. Also refers to static blocks in classes (since
     * they are valid when and only when methods are).
     * <p>
     * Can be contained only by a {@link #CLASS} scope; can contain
     * {@link #CLASS}, {@link #BLOCK}, and {@link #COMMENT} scopes.
     */
    METHOD {
        boolean canContain(State state) {
            switch (state) {
                case CLASS:
                case BLOCK:
                case COMMENT:
                    return true;
                default:
                    return false;
            }
        }
    },
    /**
     * Anything within curly braces in the body of a method.
     * <p>
     * Can be contained either by a {@link #METHOD} scope or by another
     * {@link #BLOCK} scope. Can contain {@link #CLASS}, {@link #BLOCK},
     * and {@link #COMMENT} scopes. 
     */
    BLOCK {
        boolean canContain(State state) {
            switch (state) {
                case CLASS:
                case BLOCK:
                case COMMENT:
                    return true;
                default:
                    return false;
            }
        }
    },
    /**
     * Text within a multiline comment (either JavaDoc or plain).
     * <p>
     * Cannot contain other scopes.
     */
    COMMENT {
        boolean canContain(State state) {
            return false;
        }
        
        @Override
        String indent(String line) {
            return " * " + line;
        }
    }
    ;
    /**
     * Determine whether this state can transition to the given state. So
     * called because this determines what can contain what; for instance, a
     * class can contain a method, so the {@link #CLASS} state can transition
     * to {@link #METHOD} and {@code CLASS.canContain(METHOD) == true}.
     * @param state The state to test for transition to.
     * @return {@code true} if this state can transition to the given state.
     */
    abstract boolean canContain(State state);
    
    String indent(String line) {
        return Scope.INDENTATION + line;
    }
}