/** @author Luke Maurer */
package parser;

import java.util.*;

class Block {
    final List<Widget> children;
    final Map<String, List<String>> attrs;
    final Map<String, Integer> attrLineNums;

    private static ParseException mkParseExceptionWithMsg(GUINode node,
            final String msg) {
        final ParseException e = node.parser.generateParseException();

        ParseException f = new ParseException(e.currentToken,
                e.expectedTokenSequences, e.tokenImage) {
            @Override
            public String getMessage() {
                return msg;
            }
        };

        return f;
    }

    void fillWithStmts(List<GUINode> stmts, types.Widget parent,
                types.GUIFile file) 
            throws ParseException {
        final Map<String, Widget> namedChildren = new TreeMap<String, Widget>();
        final List<Widget> anonChildren = new LinkedList<Widget>();
        
        for (GUINode stmt : stmts) {
            switch (stmt.kind) {
            case WIDGETSTATEMENT:
                for (Widget child : Widget.fromNode(stmt, parent, file))
                    if (child.id.equals("")) {
                        anonChildren.add(child);
                        children.add(child);
                    } else if (namedChildren.containsKey(child.id)) {
                        final ParseException e = mkParseExceptionWithMsg(stmt,
                                "duplicate identifier " + child.id);
                        e.fillInStackTrace();
                        throw e;
                    } else {
                        namedChildren.put(child.id, child);
                        children.add(child);
                    }
                break;
            case ACCUMSTATEMENT:
                for (Widget child : Widget.fromNode(stmt, parent, file))
                    if (child.id == null)
                        throw new MalformedASTException(stmt);
                    else if (!namedChildren.containsKey(child.id)) {
                        final ParseException e = mkParseExceptionWithMsg(stmt,
                                "undeclared identifier " + child.id);
                        e.fillInStackTrace();
                        throw e;
                    } else
                        namedChildren.get(child.id).absorb(child);
                break;
            case ATTRSTATEMENT:
                final List<GUINode> attrParts = stmt.getChildren();
                final String attr = attrParts.get(0).getText();
                final int line = attrParts.get(0).getLineNum();
                final List<String> values = new ArrayList<String>();
                for (GUINode part : attrParts.subList(1, attrParts.size()))
                    values.add(part.getText());
                attrs.put(attr, values);
                attrLineNums.put(attr, line);
                break;
            default:
                throw new MalformedASTException(stmt);
            }
        }

    }
    
    static Block blockFromStmts(List<GUINode> stmts, types.Widget parent,
                types.GUIFile file)
            throws ParseException {
        final Block block = new Block();
        block.fillWithStmts(stmts, parent, file);

        return block;
    }
    
    Block(List<Widget> children,
            Map<String, List<String>> attrs,
            Map<String, Integer> attrLineNums) {
        this.children = children;
        this.attrs = attrs;
        this.attrLineNums = attrLineNums;
    }
    
    Block() {
        this(new LinkedList<Widget>(), new TreeMap<String, List<String>>(),
                new TreeMap<String, Integer>());
    }
}