package parser;

import java.io.File;
import java.util.*;

import errors.BadAttributeNameException;
import errors.BadAttributeValueException;
import errors.BadOptionNameException;
import errors.BadOptionValueException;
import errors.GUIFileException;

import types.ErrorHandler;
import types.GUIAttributes;
import util.Builders;

final class GUIFile extends types.GUIFile implements ParserTreeConstants {
    private GUIFile(String filename, List<Widget> widgets, Set<Flag> flags,
            List<EventHandler> handlers) {
        super(filename, widgets, flags, handlers);
    }
    
    private static final Map<String, List<String>> optionDefaults =
        Builders.buildMultiMap(new TreeMap<String, List<String>>())
            .putForKeys(null, 
                    Flag.keywordTable.keySet().toArray(new String[0]))
            .put("projectName", (String) null)
            .put("controllerId", "Controller")
            .put("packageId", (String) null)
            .finish();
   
    private String projectName, packageId, controllerId;
    
    static GUIFile fromNode(String filename, GUINode node,
                ErrorHandler errorHandler)
            throws ParseException, GUIFileException {
        assert node.kind == Kind.GUIFILE;
        
        // There's a circular relationship between Widgets and GUIFiles, so we
        // need to declare the GUIFile and its members now and fill them in
        // as we go.
        final List<EventHandler> handlers = new ArrayList<EventHandler>();
        final Set<Flag> flags = new TreeSet<Flag>();
        final Block block = new Block();
        final GUIFile file = new GUIFile(filename, block.children, flags,
                handlers);

        // Find the non-special statements; they can be processed just as if
        // they weren't at the top level
        
        final List<GUINode> normalStmts = new ArrayList<GUINode>();
        final GUINode controllerBlock;
        {
            GUINode cBlock = null;
            for (GUINode child : node.getChildren())
                switch (child.kind) {
                case ATTRSTATEMENT:
                case WIDGETSTATEMENT:
                case ACCUMSTATEMENT:
                    normalStmts.add(child); break;
                case CONTROLLERBLOCK:
                    cBlock = child; break;
                default:
                }
            controllerBlock = cBlock;
        }
        
        processControllerBlock(controllerBlock, handlers);
        block.fillWithStmts(normalStmts, null, file);
        
        // XXX This is an ugly, ugly kludge.
        // We need to make a dummy Widget and fill its attrLineNumbers map with
        // dummy entries in order to avoid null pointer exceptions in exception
        // constructors. Ewwwww.
        final types.Widget dummyWidget = 
            new types.Widget(null, null, file, -1) { };
        for (String attr : block.attrs.keySet())
            dummyWidget.attrLineNums.put(attr, -1);
        
        final GUIAttributes attrs =
            new GUIAttributes(optionDefaults, dummyWidget);
        try {
            attrs.putAllAttributes(block.attrs);
        } catch (BadAttributeNameException e) {
            errorHandler.handleError(new BadOptionNameException(
                    block.attrLineNums.get(e.attrName), e.attrName));
        }
        for (String attr : attrs.keySet())
            try {
                if (Flag.keywordTable.containsKey(attr)
                        && attrs.getBoolean(attr) == true)
                    flags.add(Flag.keywordTable.get(attr));
            } catch (BadAttributeNameException e) {
                errorHandler.handleError(new BadOptionNameException(
                        block.attrLineNums.get(e.attrName), e.attrName));
            } catch (BadAttributeValueException e) {
                errorHandler.handleError(new BadOptionValueException(
                        e.getMessage(), block.attrLineNums.get(e.attrName), 
                        e.attrName, e.attrValue));
            }
        
        if (flags.contains(Flag.MOCKUP)) {
            flags.add(Flag.SUPPRESS_MODIFIERS);
            flags.add(Flag.GENERATE_MAIN);
        }
        
        {
            final String projectName = attrs.getString("projectName");
            file.projectName = projectName != null ? projectName :
                    // Default to project name based on filename
                    filenameToProjectName(filename);
        }
        
        file.controllerId = attrs.getString("controllerId");
        file.packageId = attrs.getString("packageId");
        
        return file;
    }
    
    private static String filenameToProjectName(String filename) {
        // Get rid of the directory path and the extension and make the
        // filename into a valid Java identifier
        return new File(filename).getName().replaceFirst("\\..*$", "")
                .replaceFirst("$\\P{javaJavaIdentifierStart}", "")
                .replaceAll("\\P{javaJavaIdentifierPart}", "_");
    }
    
    private static void processControllerBlock(GUINode node,
            List<? super EventHandler> tgt) {
        if (node == null)
            return;
        assert node.kind == Kind.CONTROLLERBLOCK;
        for (GUINode stmt : node.getChildren()) {
            if (stmt.kind != Kind.HANDLERSTATEMENT)
                throw new MalformedASTException(node);
            final List<GUINode> ids = stmt.getChildren();
            final String handlerType = ids.get(0).getText();
            for (GUINode handlerId : ids.subList(1, ids.size()))
                tgt.add(new EventHandler(handlerType, handlerId.getText()));
        }
    }

    public String getControllerId() {
        return controllerId;
    }

    protected void setControllerId(String controllerId) {
        this.controllerId = controllerId;
    }

    public String getPackageId() {
        return packageId;
    }

    public void setPackageId(String packageId) {
        this.packageId = packageId;
    }

    public String getProjectName() {
        return projectName;
    }

    // Changed access to public to allow command line to override projectName,
    // similar to the way the command line flags override file defaults.
    public void setProjectName(String projectName) {
        this.projectName = projectName;
    }
}