package parser;

import java.io.InputStream;
import java.util.*;

import types.ErrorHandler;

import errors.*;

/** Base class for the generated parser. Implements the functionality required
 *  by the Parser.jjt grammar file. */
abstract class ParserBase implements ParserConstants {
    /** Quick command-line test functionality; reads from standard input and
     *  displays a syntax tree. 
     * @param args Command-line arguments; ignored.
     * @throws Throwable If bad things happen. */
    public static void main(String ... args) throws Throwable {
        Parser parser = new Parser(System.in);
        GUINode guiFile = parser.GUIFile();
        System.out.println(guiFile.display());
    }

    /** The filename (possibly fake) of the file being parse. */
    protected String filename;
    
    /** The scope stack is simply a stack of scopes, a scope being the set of
     *  all identifiers declared so far. This is used to distinguish
     *  accumulation from declaration.
     *
     *  Conceptually, this is not the entire stack; rather, scopeIds is the top
     *  of the stack, and encScopeIds is the next one down. This relationship
     *  is an invariant that must be maintained. */
    protected final Stack<TreeSet<String>> scopeStack;
    
    /** The ids in the current scope, that is, the declaration/accumulation
     *  now being parsed. When the current node scope is closed, this will
     *  disappear. */
    protected TreeSet<String> scopeIds;
    
    /** The ids in the currently enclosing scope. These are the ones that are
     *  in scope from one declaration block to the next. Any declarator that
     *  includes an identifier adds this id to the enclosing scope. When the
     *  node scope is closed, this will be assigned to scopeIds and a scope
     *  will be popped off the stack. */
    protected TreeSet<String> encScopeIds;
    
    /**
     * The JJTree hook called when a node scope is opened.
     * @param node The opened node.
     */
    protected final void jjtreeOpenNodeScope(GUINode node) {
        // Remember what the first token of the node is
        node.setFirstToken(node.parser.getToken(1));

        // Advance the scope stack
        scopeStack.push(encScopeIds);
        encScopeIds = scopeIds;
        scopeIds = new TreeSet<String>();
    }

    /**
     * The JJTree hook called when a node scope is closed.
     * @param node The closed node.
     */
    protected final void jjtreeCloseNodeScope(GUINode node) {
        // Remember what the last token of the node is
        node.setLastToken(node.parser.getToken(0));

        // This scope is done; return to the next one higher
        scopeIds = encScopeIds;
        encScopeIds = scopeStack.pop();
    }

    /** The {@link ErrorHandler} to use for reporting errors. */
    protected ErrorHandler errorHandler = new ErrorHandler.Trivial();
    
    /** Parse an input stream into a list of widgets.
     *  @param ins Any input stream, to be parsed.
     *  @param filename The filename associated with the input stream. May be
     *      faked (e.g. "&lt;stdin>") if no true filename is known.
     *  @param handler An error handler.
     *  @return A GUIFile object, as described by the input, or
     *      <code>null</code> if there were errors and the handler handled all
     *      of them.
     *  @throws errors.GUIFileException If the input is invalid and the error
     *      handler couldn't handle an error. */
    public static types.GUIFile parse(InputStream ins, String filename,
                ErrorHandler handler)
            throws GUIFileException {
        final Parser parser = new Parser(ins);
        parser.filename = filename;
        if (handler != null)
            parser.errorHandler = handler;
        try {
            final GUIFile ans = GUIFile.fromNode(filename, parser.GUIFile(), 
                parser.errorHandler);
            if (parser.errorHandler.anyErrors())
                // There were errors; the GUI file is invalid
                return null;
            return ans;
        } catch (ParseException e) {
            try {
                parser.raiseError(e);
                return null;
            } catch (ParseException e2) {
                throw new SyntaxException(e2);
            }
//        } catch (TokenMgrError e) {
//            try {
//                parser.raiseError(e);
//                return null;
//            } catch (TokenMgrError e2) {
//                throw new SyntaxException(e2);
//            }
        }
    }
    
    /** Parse an input stream into a list of widgets.
     *  @param ins Any input stream, to be parsed.
     *  @return A GUIFile object, as described by the input.
     *  @throws errors.GUIFileException If the input is invalid. */
    public static types.GUIFile parse(InputStream ins) 
            throws GUIFileException {
        return parse(ins, "<unknown>", null);
    }
    
    /**
     * Skip parsing until a token of one of the given types is found.
     * @param skipTgt A token type to stop skipping on.
     * @param moreTgts Other token types.
     * @return The first token found matching one of the given types, or the
     *         EOF token.
     */
    protected final Token skipUntil(int skipTgt, Integer ... moreTgts) {
        java.util.Set<Integer> tgts = new java.util.TreeSet<Integer>();
        java.util.Collections.addAll(tgts, moreTgts);
        tgts.add(skipTgt);
        tgts.add(EOF);
        Token tok;
        do {
            tok = getNextToken();
        } while (!tgts.contains(tok.kind));
        return tok;
    }
    
    /**
     * Report the given error using the error handler.
     * @param error The exception representing the error to report.
     * @throws ParseException If the error handler passes on the error.
     */
    protected final void raiseError(ParseException error) 
            throws ParseException {
        final SyntaxException handledError = new SyntaxException(error);
        if (!errorHandler.handleError(handledError))
            // Throw the original error
            throw error;
    }
    
    /**
     * Report a syntax error with the given error message using the error 
     * handler.
     * @param msg The error message.
     * @throws ParseException If the error handler passes on the error.
     */
    protected final void raiseError(String msg) throws ParseException {
        final SyntaxException handledError = new SyntaxException(msg,
                generateParseException());
        if (!errorHandler.handleError(handledError))
            // Throw a basic parse error with the message
            throw new ParseException(msg);
    }
    
    /**
     * Mark the current block, so that parsing may skip to the end.
     */
    protected final void markBlock() {
        getTokenSource().markBlock();
    }
    
    private void skipMarkedBlock() {
        final int tgtLevel = getTokenSource().encLevelStack.peek();
        if (tgtLevel <= 0)
            return;
        do {
            getNextToken();
        } while (getTokenSource().encLevel >= tgtLevel);
    }

    /*
    private void skipToMarkedBlock() {
        final int tgtLevel = getTokenSource().encLevelStack.peek();
        if (tgtLevel <= 0)
            return;
        do {
            getNextToken();
        } while (getTokenSource().encLevel > tgtLevel);
    }
    */

    /** Skip the current statement and continue the block. */
    protected final void continueBlock() {
        /* If there's an open brace before the next semicolon, we'll need to
         * deal with the enclosed block. If there's a semicolon first, we can
         * just skip through the semicolon. */
        Token tok = skipUntil(BRACE_OPEN, BRACE_CLOSE, SEMICOLON);

        if (tok.kind == BRACE_OPEN) {
            markBlock();
            skipMarkedBlock();
        }
    }
    
    /** Skip to the end of the current block. */
    protected void leaveBlock() {
        markBlock();
        skipMarkedBlock();
    }
    
    /**
     * Get the next token parsed. Implemented by the {@link Parser} class
     * generated by JavaCC.
     * @return The next token parsed.
     */
    protected abstract Token getNextToken();
    
    /**
     * Make a parse exception and pass it the current state of the parser.
     * Implemented by the {@link Parser} class generated by JavaCC.
     * @return A new {@link ParseException}.
     */
    protected abstract ParseException generateParseException();

    /**
     * Retrieve the lexer (known to JavaCC as the token manager). Implemented
     * by the {@link Parser} class generated by JavaCC.
     * @return The lexer being used for parsing.
     */
    protected abstract ParserTokenManager getTokenSource();
    
    protected ParserBase() {
        filename = "<unknown>";
        scopeStack = new Stack<TreeSet<String>>();
        scopeIds = new TreeSet<String>();
        encScopeIds = new TreeSet<String>();
        scopeStack.push(encScopeIds);
        scopeStack.push(scopeIds);
    }
}
