package parser;

import java.util.*;

import util.Pair;

/** The class of widgets returned by the Parse.parse() method. To other
 *  packages, these look and act like types.Widget objects.
 *  
 *  <p>The public Widget class, to be used globally, is types.Widget. This
 *	class is used to construct instances of that class. */
class Widget extends types.Widget implements ParserTreeConstants {
    /** The syntax node that created this widget. */
    final GUINode origin;
    
    /** @param type The identifier of the type of this widget.
     *  @param id The identifier of this widget.
     *  @param file The containing GUIFile.
     *  @param lineNum the line number on which this widget begins.
     *  @param origin The node describing this widget.
     *  @deprecated Use {@link #Widget(String, String, types.Widget,
     *      types.GUIFile, int, GUINode)} and supply the parent widget. */
    Widget(String type, String id, types.GUIFile file,
            int lineNum, GUINode origin) {
        super(type, id, file, lineNum); 
        this.origin = origin;
    }
    
    /** @param type The identifier of the type of this widget.
     *  @param id The identifier of this widget.
     *  @param parent The widget's parent widget, or <code>null</code> for a
     *      top-level widget.
     *  @param file The containing GUIFile.
     *  @param lineNum the line number on which this widget begins.
     *  @param origin The node describing this widget. */
    Widget(String type, String id, types.Widget parent, types.GUIFile file,
            int lineNum, GUINode origin) {
        super(type, id, parent, file, lineNum); 
        this.origin = origin;
    }
    
    /**
     * Constructs a list of widgets from a single syntax node.
     * @param node The node from which to build the widgets.
     * @param parent The widgets' parent, already built, or {@code null} for
     *      top-level widgets.
     * @param file The file under construction.
     * @return The widgets described by the node.
     * @throws ParseException If the node describes invalid input.
     * @throws MalformedASTException If something goes very, very wrong; i.e.
     *      the parser is buggy.
     */
    static List<Widget> fromNode(GUINode node, types.Widget parent,
                types.GUIFile file)
            throws MalformedASTException, ParseException {
        
        switch (node.kind) {
            /* case GUIFILE: {
                List<Widget> ans = new LinkedList<Widget>();
                for (GUINode child : node)
                    ans.addAll(fromNode(child));
                return ans;
            } */

            case WIDGETSTATEMENT:
            /* case ISOLIDSTATEMENT: */ {
                final String type;
                final int lineNum;
                final List<Pair<String, String>> ids =
                    new ArrayList<Pair<String, String>>();

                final List<GUINode> stmtParts = node.getChildren();
                final int nParts = stmtParts.size();

                // The statement has to have something
                if (nParts == 0)
                    throw new MalformedASTException(node);
                
                // The statement begins with a type identifier
                if (stmtParts.get(0).kind != Kind.IDENTIFIER)
                    throw new MalformedASTException(node);
                
                type = stmtParts.get(0).getText();
                lineNum = stmtParts.get(0).getLineNum();
                
                /* There are zero or more widget declarators, followed by zero
                 * or more statements. Find out where the divide is. */

                List<GUINode> decls, stmts;
                {
                    Pair<? extends List<GUINode>, ? extends List<GUINode>>
                        split = GUINode.splitDeclsFromStmts(
                                stmtParts.subList(1, nParts));
                    decls = split.a;
                    stmts = split.b;
                }

                if (decls.isEmpty()) {
                    // There's a single anonymous widget being declared
                    ids.add(Pair.pair("", (String) null));
                }
                
                for (GUINode decl : decls) {
                    List<GUINode> declParts = decl.getChildren();
                    switch (declParts.size()) {
                        case 1:
                            // Either an identifier or a label
                            switch (declParts.get(0).kind) {
                                case IDENTIFIER:
                                    ids.add(Pair.pair(
                                                declParts.get(0).getText(),
                                                (String) null));
                                    break;
                                case STRINGVALUE:
                                case BARESTRING:
                                    ids.add(Pair.pair(
                                                "",
                                                declParts.get(0).getText()));
                                    break;
                                default:
                                    throw new 
                                        MalformedASTException(decl);
                            }
                            break;
                        case 2:
                            // Identifier with label
                            ids.add(Pair.pair(
                                        declParts.get(0).getText(),
                                        declParts.get(1).getText()));
                            break;
                        default:
                            throw new MalformedASTException(decl);
                    }
                }

                final List<Widget> ans = new ArrayList<Widget>();

                for (Pair<String, String> idAndLabel : ids) {
                    String id = idAndLabel.a, label = idAndLabel.b;
                    Widget newWidget = new Widget(type, id, parent, file,
                            lineNum, node);
                    Block block = Block.blockFromStmts(stmts, newWidget, file);
                    
                    if (label != null) {
                        newWidget.attrs.put("label",
                                Collections.singletonList(label));
                    }

                    newWidget.attrs.putAll(block.attrs);
                    newWidget.attrLineNums.putAll(block.attrLineNums);
                    newWidget.children.addAll(block.children);
                    
                    ans.add(newWidget);
                }

                return ans;
            }

            case ACCUMSTATEMENT: {
                final List<String> ids = new ArrayList<String>();
                final List<GUINode> stmtParts = node.getChildren();
                final int nParts = stmtParts.size();

                if (nParts == 0)
                    throw new MalformedASTException(node);

                List<GUINode> decls, stmts;
                {
                    Pair<? extends List<GUINode>, ? extends List<GUINode>>
                        split = GUINode.splitDeclsFromStmts(stmtParts);
                    decls = split.a;
                    stmts = split.b;
                }
                
                for (GUINode decl : decls) {
                    if (decl.kind != Kind.WIDGETDECLARATOR)
                        throw new MalformedASTException(node);
                    ids.add(decl.getChildren().get(0).getText());
                }

                Block block = Block.blockFromStmts(stmts, parent, file);

                final List<Widget> ans = new ArrayList<Widget>();

                for (String id : ids) {
                    Widget newWidget = new Widget(null, id, parent, file,
                            node.getLineNum(), node);
                    newWidget.attrs.putAll(block.attrs);
                    newWidget.children.addAll(block.children);
                    ans.add(newWidget);
                }

                return ans;
            }

            default:
                throw new MalformedASTException(node);
        }
    }

    void absorb(Widget widget) {
        children.addAll(widget.children);
        attrs.putAll(widget.attrs);
    }
}
