package types;

import java.util.*;

import static util.Functional.*;

import errors.BadAttributeException;
import errors.BadChildWidgetException;
import errors.BadLanguageDefinitionException;
import errors.BadWidgetTypeException;
import errors.GUIFileException;


/**
 *  LangDef is the abstract class that all Language Definitions
 * inherit from.  It contains several utility methods and one
 * method stub, generateCode, which is called by the Controller
 * and returns a double array of strings, with the first entry
 * corresponding to the file name to be outputted to, and the
 * second entry corresponding to the contents of that file.
 */

public abstract class LangDef {
	/** The pathname to the language's lib folder. */
	protected final String libPath;
	
	/**
     * A running list of all the distinct import statements required by various
     * Widgets within the code.
     */
	protected final Set<String> imports;
	
	/**
     * A running list of all the distinct modifiers to be included in the final
     * generated code.
     */
	protected final Set<String> modifiers;
	
    /**
     * A running list of all the distinct fields to be included in the final
     * generated code.
     */
    protected final Set<String> variables;
	
	/**
     * A running list of all the distinct inner classes to be included in the
     * final generated code.
     */
	protected final Set<String> innerClasses;
	
    /**
     * The {@link ErrorHandler} to use for reporting errors.
     */
    protected ErrorHandler errorHandler = new ErrorHandler.Trivial();
    
	/**
	 * @param libPath The package prefix for the widget library.
	 * @param imports The set to use for {@link #imports}.
	 * @param modifiers The set to use for {@link #modifiers}.
	 * @param innerClasses The set to use for {@link #innerClasses}.
	 * @param variables The set to use for {@link #variables}.
	 */
	protected LangDef(String libPath, Set<String> imports,
            Set<String> modifiers, Set<String> innerClasses,
            Set<String> variables) {
        this.libPath = libPath;
        this.imports = imports;
        this.variables = variables;
        this.modifiers = modifiers;
        this.innerClasses = innerClasses;
    }
    
    /**
     * @param libPath The package prefix for the widget library.
     */
    protected LangDef(String libPath) {
        this(libPath, new HashSet<String>(), new HashSet<String>(),
                new HashSet<String>(), new HashSet<String>());
    }

    /**
	 * generateCode accepts the generated Parser output and manipulates it
	 * according to the structure of the language to generate the appropriate
	 * output code.
	 * @param guiFile A GUIFile object, as returned by the parser.
     * @param errorHandler The {@link ErrorHandler} object to use for reporting
     *      errors.
	 * @return An array of two-element arrays. The first element should be a
     *      filename, and the second element should have the contents of the
     *      generated file with that name.
     * @throws GUIFileException If any errors are found in the file.
	 */
	abstract public String[][] generateCode(
            GUIFile guiFile, ErrorHandler errorHandler) 
	            throws GUIFileException;
    
    /**
     * @return The widget type to use as the base of the widget hierarchy. This
     *         will typically be "Widget".
     */
    abstract public Class<? extends WidgetDef> getRootWidgetType(); 

	/**
	 *     createWidget is a utility method included to cut down on
	 *  potential code duplication.  It is not necessary to use
	 *  or override it if the Language Definition does not desire to.  
	 *
	 *    createWidget consults with the library definitions of the
	 * widget type to assemble the code necessary for declaring a 
	 * widget in the given language.  If structural constraints of
	 * the language cause this method to be ineffective, it can be
	 * overridden in the subclass. It currently recursively generates
	 * child widgets as they are encountered.
	 * 
     * @param widget The widget from the {@link GUIFile} to create.
     * @return The code to create the widget.
	 * @throws BadAttributeException If an invalid attribute is given.
	 * @throws BadWidgetTypeException If the widget type does not exist.
	 * @throws BadChildWidgetException For an incompatible child type. 
	 */
	protected String createWidget(Widget widget) throws GUIFileException {
        // In the base case, call the two-argument version and pass in no
        // child attribute defaults.
		return createWidget(widget, null);
	}
    
    /**
     * Create a widget which may have child attributes granted by its parent
     * widget.
     * @param widget The widget to generate code for.
     * @param parentDef The widget definition for the widget's parent, or
     *        {@code null}; if {@code null} then do not allow any child
     *        attributes.
     * @return Generated code to create the widget and its children.
     * @see WidgetDef#getChildDefaults()
     */
    protected String createWidget(Widget widget, WidgetDef parentDef)
            throws GUIFileException {
        final StringBuilder builder = new StringBuilder();
        final WidgetDef widgetDef = 
            WidgetDef.defineWidget(this, widget, parentDef);
        
        // The processAttributes method will take care of base widget types.
        final GUIAttributes attrs;
        
        try {
            attrs = widgetDef.processAttributes();
        } catch (BadAttributeException e) {
            if (!errorHandler.handleError(e))
                throw e;
            // We can't do anything without a valid GUIAttributes object, so
            // bail now (nothing's going to be output anyway)
            return "";
        }
        
        final StringBuilder attrBuilder = new StringBuilder();
        
        // Give the widget def and all its base widget types a chance to add
        // their attributes.
        for (WidgetDef ancestor : backward(widgetDef.getAncestry()))
            try {
                attrBuilder.append(ancestor.addAttributes(attrs));
            } catch (BadAttributeException e) {
                if (!errorHandler.handleError(e))
                    throw e;
            }
        
        variables.addAll(widgetDef.getVariables());
        
        for (WidgetDef ancestor : backward(widgetDef.getAncestry())) {
            modifiers.addAll(ancestor.getModifiers());
            innerClasses.addAll(ancestor.getInnerClasses());
            if (ancestor.isAbstract()) {
	            	variables.addAll(ancestor.getVariables());
	            	builder.append(ancestor.preamble());
            }
        }
        
        builder.append(widgetDef.preamble());
        
        // Attribute code is generated first, to ensure the widget is initialized.
        // But the code is added here, after the preamble.
        builder.append(attrBuilder);
        
        for (Widget child : widget.children) {
            final WidgetDef childDef;
            try {
                final String childCode = createWidget(child, widgetDef);
                if (childCode.equals(""))
                    // There was an error; bail
                    continue;
                
                builder.append(childCode);
                childDef =
                    WidgetDef.defineWidget(this, child, widgetDef);
                final GUIAttributes childAttrs = 
                    childDef.processAttributes();
                
                for (WidgetDef ancestor : backward(widgetDef.getAncestry())) 
                    if (ancestor.isAbstract())
                        try {
                            builder.append(ancestor.addChild(
                                    child.type, child.id, childAttrs));
                        } catch (BadChildWidgetException e) {
                            // If a particular abstract base doesn't support
                            // adding child widgets, we don't really care. All
                            // that has to matter is the concrete type.
                            continue;
                        }
                
                builder.append(widgetDef.addChild(
                        child.type, child.id, childAttrs));
            } catch (GUIFileException e) {
                if (!errorHandler.handleError(e))
                    throw e;
            }
        }
    
        for (WidgetDef ancestor: backward(widgetDef.getAncestry())) {
            imports.addAll(ancestor.getImports());
            builder.append(ancestor.postamble());
        }
        return builder.toString();
    }

    /**
     * @return The fully qualified name of the package containing the
     *      WidgetDefs for this language.
     */
    public String getLibPath() {
        return libPath;
    }

    /**
     * Get a LangDef instance for the specified language.
     * @param lang The name of the language, for instance "Java".
     * @return The corresponding LangDef.
     */
    public static LangDef defineLanguage(String lang) {	
    	try {
    		Class langDef;
    		langDef = Class.forName("lib.lang."+lang);
    		return (LangDef) langDef.newInstance();
    	} catch (ClassNotFoundException e) {
    		throw new BadLanguageDefinitionException(lang);
    	} catch (InstantiationException e) {
    		throw new BadLanguageDefinitionException(lang);
    	} catch (IllegalAccessException e) {
    		throw new BadLanguageDefinitionException(lang);
    	}
    }
}
