package types;

import java.util.*;

import util.Functional;
import util.Output;

/** The general type of all widgets. */
public abstract class Widget {
    /** The identifier of the widget. */
    public final String id;
    /** The identifier of the type of the widget. */
    public final String type;
    /** The widget's parent, or <code>null</code> for a top-level widget. */
    public final Widget parent;
    /** A list of the widget's children. Is probably not modifiable. */
    public final List<Widget> children;
    /** A mapping from attribute names to values. Is probably not modifiable.*/
    public final Map<String, List<String>> attrs;
    
    /** The .gui file declaring this widget. */
    public final GUIFile file;
    /** The line number at which this widget was declared. */
    public final int lineNum;
    /** A mapping from attribute names to the line numbers at which they were
     *  specified. */
    public final Map<String, Integer> attrLineNums;

    /** The number appended to the last anonymous identifier generated. */
    protected static int lastAnonIdNumber = 0;
    
    /** @return A unique, newly generated anonymous identifier. */
    protected static String mkAnonId() {
        return String.format("anon$%d", lastAnonIdNumber++);
    }
    
    /**
     * Make an anonymous identifier if needed.
     * @param id An identifier.
     * @return {@code id} if {@code id != null && id.length() > 0}; a new
     *         anonymous identifier otherwise.
     * @see Widget#mkAnonId()
     */
    protected static String mkIdIfNeeded(String id) {
        if (id == null || id.length() == 0)
            return mkAnonId();
        return id;
    }
    
    /**
     * Make an anonymous identifier for a widget of the given type.
     * @param type The type of the widget.
     * @return A unique, newly generated anonymous identifier.
     */
    protected static String mkAnonId(String type) {
        return String.format("anon$%s$%d", type, lastAnonIdNumber++);
    }
    
    /**
     * Make an anonymous identifier, if needed, for a widget of the given type.
     * @param id An identifier.
     * @param type The type of the widget.
     * @return {@code id} if {@code id != null && id.length() > 0}; a new
     *         anonymous identifier otherwise.
     * @see Widget#mkAnonId()
     */
    protected static String mkIdIfNeeded(String id, String type) {
        if (id == null || id.length() == 0)
            return mkAnonId(type);
        return id;
    }

    /** @param type The identifier of the type of the widget.
     *  @param id The identifier of the widget, or <code>null</code> for
     *      anonymous.
     *  @param file The GUIFile in which the widget was declared.
     *  @param lineNum The line number at which this widget was declared. */
    protected Widget(String type, String id, GUIFile file, int lineNum) {
        this(type, id, null, file, lineNum);
    }

    /** @param type The identifier of the type of the widget.
     *  @param id The identifier of the widget, or <code>null</code> for
     *      anonymous.
     *  @param parent The widget's parent widget, or <code>null</code> for a
     *      top-level widget.
     *  @param file The GUIFile in which the widget was declared.
     *  @param lineNum The line number at which this widget was declared. */
    protected Widget(String type, String id, Widget parent, GUIFile file,
            int lineNum) {
        this.id = mkIdIfNeeded(id, type);
        this.type = type;
        this.parent = parent;
        this.file = file;
        this.lineNum = lineNum;
        children = new ArrayList<Widget>();
        attrs = new TreeMap<String, List<String>>();
        attrLineNums = new TreeMap<String, Integer>();
    }
    
    /**
     * Format a list of values for printing.
     * @param values A list of values.
     * @return The values, surrounded by typographer's (Unicode) quotes and
     *         delimited by commas.
     */
    protected static String formatValueList(List<String> values) {
        return Functional.join(Functional.wrapEach(
                values, "\u201c", "\u201d"), ", ");
    }
   
    public String toString() {
        StringBuilder builder = new StringBuilder(
                String.format("Widget: %s “%s”\n", type, id));
        if (parent != null)
            builder.append(Output.indent(
                    String.format("Parent: %s\n", parent.id)));
        if (!attrs.isEmpty()) {
            builder.append(Output.indent("Attributes:\n"));
            for (Map.Entry<String, List<String>> attr : attrs.entrySet())
                builder.append(Output.indent(Output.indent(
                        String.format("%s: %s\n",
                                attr.getKey(), 
                                formatValueList(attr.getValue())))));
            }
        if (!children.isEmpty()) {
            builder.append(Output.indent("Children:\n"));
            for (Widget child : children)
                builder.append(Output.indent(Output.indent(child.toString())));
        }

        return builder.toString();
    }
}
