package types;

import java.lang.annotation.*;
import java.lang.reflect.*;
import java.util.*;

import util.Triple;

import static util.Functional.*;

import errors.BadAttributeException;
import errors.BadChildWidgetException;
import errors.BadWidgetDefinitionException;
import errors.BadWidgetTypeException;

/**
 *    WidgetDef is the abstract class that all definitions in the library
 * implement.
 * 
 * Standardized Formats for our Libraries:
 * 
 * String Concatenation: Use the StringBuilder class and the 
 * StringBuilder.append() method to concatenate strings.
 * 
 * String Formatting: use the String.format(String, args) to
 * format strings.
 * 
 * Use Window.java class as an example.
 *
 */
public abstract class WidgetDef {

    /** Widget object that is passed as a parameter in the constructor. */
	final protected Widget widget;
	/** id of Widget object */
	final public String id;
	/** Parent widget def */
    final protected WidgetDef parent;    
    
    /* The LangDef that defined this Widget */
    private LangDef langDef = null;
    
	/**
     * @param widget The Widget object.
     * @param parent The WidgetDef for the parent widget, or {@code null} for a
     *        top-level widget.
     */
	protected WidgetDef(Widget widget, WidgetDef parent) {
		this.widget = widget;
        this.parent = parent;
		id = widget.id;
	}

    /**
     * @param widget The Widget object.
     * @param parent The WidgetDef for the parent widget, or {@code null} for a
     *        top-level widget.
     * @param langDef The language definition in use.
     */
    protected WidgetDef(Widget widget, WidgetDef parent, LangDef langDef) {
        this.widget = widget;
        this.parent = parent;
        this.langDef = langDef;
        id = widget.id;
    }
    
    /**
     * Format the widget identifier for use in comments. Replaces the id of an
     * anonymous widget with something more readable.
     * @return {@code id}, if the widget is not anonymous; a placeholder
     * otherwise.
     */
    protected String getCommentName() {
        return commentForm(id, widget.type);
    }
    
    /**
     * Format a widget identifier for use in comments. Replaces the id of an
     * anonymous widget with something more readable.
     * @param id The identifier of the widget.
     * @param type The type of the widget.
     * @return {@code id}, if the widget is not anonymous; a placeholder
     * otherwise.
     */
    protected static String commentForm(String id, String type) {
        if (id.startsWith("anon$"))
            return String.format("(anonymous %s)", type);
        else
            return id;
    }
    
	/**
	 * Accessor method for the Widget language.
	 * @return language
	 */
	public abstract String getLanguage();

	/**
	 * Get the chunk of initialization code to be run first. Typically this
     * code will declare a local variable for the widget.
     * <p>
     * Note that, by its nature, a preamble should only be called once;
     * therefore the preamble is <i>not</i> inherited.
	 * @return Instatiation code.
	 */
	public abstract String preamble();
	
	/**
     * Get the chunk of initialization code to be run after everything else is
     * done. This is an empty string by default.
     * <p>
     * Note that the {@link #preamble()} and postamble are <i>not</i>
     * inherited (much as Java constructors aren't). 
	 * @return Final-phase initialization code.
     */
	public String postamble() {
        return "";
    }
	
	/** 
	 * Returns a chunk of code intended to declare variables which the backend
	 * code may need access to and therefore should be declared outside of the
	 * constructor code.
	 * 
	 * @return Variable Declaration for the widget.
	 */
	public abstract List<String> getVariables();
	
    private GUIAttributes guiAttributes = null;
    
    /** Create a GUIAttributes object for a widget.
     *  @return The GUIAttributes object.
     *  @throws BadAttributeException If an unrecognized attribute is listed
     *      for the widget. */
    public final GUIAttributes processAttributes() 
            throws BadAttributeException {
        
        if (guiAttributes != null)
            return guiAttributes;
            
        // Work through the inheritance tree in method resolution order
        // (see getAncestry()), adding all defaults along the way. The
        // parent widget trumps all other WidgetDefs - that is, defaults
        // from the parent widget override the entire widget hierarchy.
        
        final Map<String, List<String>> defaults =
            new TreeMap<String, List<String>>();
        final Map<String, String> synonyms =
            new TreeMap<String, String>();
        final Map<String, List<String>> multisynonyms =
            new TreeMap<String, List<String>>();
        final List<WidgetDef> ancestry = getAncestry();
        for (WidgetDef ancestor : backward(ancestry)) {
            defaults.putAll(ancestor.getDefaults());
            synonyms.putAll(ancestor.getSynonymousAttributes());
            multisynonyms.putAll(ancestor.getMultisynonymousAttributes());
        }
        if (parent != null)
            defaults.putAll(parent.getChildDefaults());
        
        guiAttributes = new GUIAttributes(
                defaults, synonyms, multisynonyms, widget);
        guiAttributes.putAllAttributes(widget.attrs);
        
        return guiAttributes;
    }
    
    /**
     * Every attribute in our language has to have a default setting.  This
     * returns all the defaults in a map of Strings to Lists of Strings.
     * @return Map of attribute defaults
     */
    public abstract Map<String,List<String>> getDefaults();

	/**
	 * The addAttributes code does two things; it creates the lib.java code
	 * necessary for the attributes and it type checks for all attributes.
	 * A BadAttributeException is thrown if the names or the values are
	 * erroneous.  If an attribute is not found in the GUIAttributes, a
	 * BadAttributeNameException is thrown from GUIAttributes.
	 * 
	 * @param attrs The attributes to set for the widget.
	 * @return The generated code to set the attributes.
	 * @throws BadAttributeException
	 */
	public abstract String addAttributes(GUIAttributes attrs)
            throws BadAttributeException;
	
    /**
     * Specifies synonymous attributes. For each entry in this map, the key is
     * the name of an attribute (which does not appear in
     * {@link #getDefaults()}), and the value is the name of the real attribute
     * with which that attribute is synonymous (i.e. equivalent).
     * <p>
     * By default, returns the empty map; WidgetDef subclasses may override
     * to specify synonymous attributes.
     * @return The map of synonymous attributes.
     */
    public Map<String, String> getSynonymousAttributes() {
        return Collections.emptyMap();
    }
    
    /**
     * Specifies multisynonymous attributes. A multisynonymous attribute is one
     * that is equivalent to giving a number of other attributes in one value
     * list. For instance, a "size" attribute to be multisynonymous with
     * "width" and "height", and then one could either give the values of width
     * and height separately or give the size attribute both values.
     * @return A mapping; each key is a multisynonymous attribute, and each
     *         value is a list of the attributes to receive its values, in the
     *         order in which they must appear in the attribute value list.
     */
    public Map<String, List<String>> getMultisynonymousAttributes() {
        return Collections.emptyMap();
    }
    
    /** Specifies the child attributes for the widget.
     * 
     *  A container widget can have associated attributes that are
     *  attributes not of the container itself but of its children. For
     *  instance, any widget in a GridLayout has attributes for its position
     *  and span; it has these attributes not because of its widget type, but
     *  because of its <i>parent's</i> widget type. Such are called <i>child
     *  attributes</i>.
     *  
     *  @return A map of the default values for the child attributes
     *  @see #getDefaults() */
    public Map<String, List<String>> getChildDefaults() {
        // Default to having no child attributes
        return Collections.emptyMap();
    }

    /**
     * Check to make sure the childType is legal for the widget, then add it.
     * @param childType The widget type of the child.
     * @param childId The identifier of the child.
     * @param attrs The child's GUIAttributes map. Child attributes will be
     *      read from this map.
     * @return The generated code to add the child to this widget.
     * @throws BadChildWidgetException If the child is of an inappropriate
     *      type. For instance, a menu can only have certain widgets (items,
     *      submenus, etc.) as children.
     * @throws BadAttributeException If one of the given attributes is
     *      erroneous. 
     */
    public String addChild(String childType, String childId, 
            GUIAttributes attrs)
            throws BadChildWidgetException, BadAttributeException {
        // Default to the pre-child-attribute behavior
        throw new BadChildWidgetException(widget);
    }

	/**
     * getModifiers returns a list of methods designed to modify various
     * attributes of the widget type in question. When getModifiers is called,
     * the accessors and mutators for all attributes are returned.
     * @return List of mutators and accessors for the widget type.
     */

	public abstract List<String> getModifiers();

    /** Analog of {@link #getModifiers()} for child attributes.
     *  @return List of mutators and accessors for attributes of children of
     *      widgets of this type. */
    public List<String> getChildModifiers() {
        // Default to having no child modifiers
        return Collections.emptyList();
    }
    
	/**
	 * getImports keeps a Set of import lines that are pertinent to the code 
	 * generated in the other methods.  
	 * @return Set of import strings.
	 */
	public abstract Set<String> getImports();
	
	/**
	 * getInnerClasses returns a set of strings corresponding to any inner 
	 * classes utilized by the widget at runtime.  This is primarily intended
	 * to make defining and manipulating (at runtime) non swing elements more
	 * straightforward and flexible.  Since the majority of definitions will 
	 * most likely not be utilizing this functionality, the default action is
	 * to return an empty set.  This can be overridden where appropriate.
	 * 
	 *  @return Set of Inner Classes.
	 */
	
	public Set<String> getInnerClasses() {
		return Collections.emptySet();
	}

    private Class<? extends WidgetDef> findWidgetClass(String name) {
        try {
            return Class.forName(langDef.getLibPath() + "." + name)
                    .asSubclass(WidgetDef.class);
        } catch (ClassNotFoundException e) {
            throw new BadWidgetDefinitionException(e);
        }
    }
    
    private static Class<? extends WidgetDef> findWidgetClass(
            String name, LangDef langDef) {
        try {
            return Class.forName(langDef.getLibPath() + "." + name)
                    .asSubclass(WidgetDef.class);
        } catch (ClassNotFoundException e) {
            throw new BadWidgetDefinitionException(e);
        }
    }
    
    private static final HashMap<
                Triple<LangDef, Class<? extends WidgetDef>, Widget>, WidgetDef> 
            widgetDefCache = 
                new HashMap<Triple<
                    LangDef, Class<? extends WidgetDef>, Widget>, WidgetDef>();
    
    /**
     * Instantiate a WidgetDef class with the given widget. Handles reflection
     * errors by wrapping them in {@link BadWidgetDefinitionException}
     * instances.
     * @param cls The class, which must extend WidgetDef and include a public
     *        constructor taking a widget and a widget definition for
     *        arguments, to instantiate.
     * @param langDef The LangDef to give the WidgetDef.
     * @param widget The widget to pass to its constructor.
     * @param parent The WidgetDef for the widget's parent, or {@code null} for
     *        a top-level widget.
     * @return A new instance of the given class.
     * @throws BadWidgetDefinitionException on any reflection error (e.g. the
     *         WidgetDef class has no suitable public constructor).
     */
    private static <C extends WidgetDef> C makeWidgetDef(
            Class<C> cls, LangDef langDef, 
            Widget widget, WidgetDef parent) {
        if (widgetDefCache.containsKey(
                Triple.of(langDef, cls, widget)))
            return cls.cast(
                    widgetDefCache.get(Triple.of(langDef, cls, widget)));
        
        try {
            final Constructor<C> constructor =
                cls.getConstructor(Widget.class, WidgetDef.class);
            final C ans = constructor.newInstance(widget, parent);
            ans.langDef = langDef;
            widgetDefCache.put(
                    Triple.<LangDef, Class<? extends WidgetDef>, Widget> 
                        of(langDef, cls, widget), ans);
            return ans;
        } catch (IllegalArgumentException e) {
            throw new BadWidgetDefinitionException(e);
        } catch (InstantiationException e) {
            throw new BadWidgetDefinitionException(e);
        } catch (IllegalAccessException e) {
            throw new BadWidgetDefinitionException(e);
        } catch (InvocationTargetException e) {
            throw new BadWidgetDefinitionException(e);
        } catch (SecurityException e) {
            throw new BadWidgetDefinitionException(e);
        } catch (NoSuchMethodException e) {
            throw new BadWidgetDefinitionException(e);
        }
    }

    private WidgetDef makeWidgetDef(Class<? extends WidgetDef> cls, 
            Widget widget, WidgetDef parent) {
        return makeWidgetDef(cls, langDef, widget, parent);
    }
    

    /**
     * Create a WidgetDef object for a given Widget and language, with the
     * given WidgetDef as the parent of the widget.
     * 
     * Calls are cached, so multiple calls with the same Widget and language
     * will return the same WidgetDef.
     * 
     * @param lang The LangDef for the language to output.
     * @param widget The Widget, as given by the parser.
     * @param parent The WidgetDef for the parent widget.
     * @return A new WidgetDef.
     * @throws BadWidgetTypeException If the Widget specifies a widget type
     *      in {@link Widget#type} that:
     *      <ul>
     *          <li>does not exist,</li>
     *          <li>is abstract, or</li>
     *          <li>is contained, when {@code widget} has no parent</li>
     *      </ul>
     */
    public static WidgetDef defineWidget(LangDef lang, Widget widget,
            WidgetDef parent) 
                throws BadWidgetTypeException {

        final Class<? extends WidgetDef> widgetDefClass;
        try {
            widgetDefClass =
                Class.forName(lang.getLibPath() + "." + widget.type)
                    .asSubclass(WidgetDef.class);
        } catch (ClassNotFoundException e) {
            throw new BadWidgetTypeException(widget);
        } catch (ClassCastException e) {
            throw new BadWidgetDefinitionException(e);
        }

        if (isAbstract(widgetDefClass))
            throw new BadWidgetTypeException(widget);
        if (parent == null && isContained(widgetDefClass, lang))
            throw new BadWidgetTypeException(
                    "widget must be enclosed in another widget", widget);

        return makeWidgetDef(widgetDefClass, lang, widget, parent);
    }

    /**
     * Create a WidgetDef object for a given Widget and language. The Widget
     * is assumed to be at the top level (i.e. it has no parent).
     * 
     * Calls are cached, so multiple calls with the same Widget and language
     * will return the same WidgetDef.
     * 
     * @param lang The LangDef for the language to output.
     * @param widget The Widget, as given by the parser.
     * @return A new WidgetDef.
     * @throws BadWidgetTypeException If the Widget specifies a widget type
     *      in {@link Widget#type} that is either abstract or nonexistant.
     */
    public static WidgetDef defineWidget(LangDef lang, Widget widget) 
                throws BadWidgetTypeException {
        return defineWidget(lang, widget, null);
    }
    /** 
     * Annotation specifying the base widget types of a widget. Note that the
     * root widget type for a language definition is always implicitly the last
     * base widget type.  
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    @Documented
    public static @interface Derived {
        /** @return The widget definition class of each parent widget type. */
        Class<? extends WidgetDef>[] value();
    }
    
    /**
     * Retrieve the base widget types.
     * @return A list of the base widget types, as WidgetDefs.
     * @see Derived
     */
    public final List<WidgetDef> getBaseWidgetDefs() {
        final Class<? extends WidgetDef> root = langDef.getRootWidgetType();
        
        final Derived annot = getClass().getAnnotation(Derived.class);
        if (annot == null)
            return Collections.singletonList(
                    makeWidgetDef(root, widget, parent));
        final Class<? extends WidgetDef>[] bases = annot.value();
        final List<WidgetDef> ans = new ArrayList<WidgetDef>(bases.length + 1);
        boolean haveRoot = false;
        for (Class<? extends WidgetDef> base : bases) {
            ans.add(makeWidgetDef(base, widget, parent));
            if (base.equals(root))
                haveRoot = true;
        }
        if (!haveRoot)
            ans.add(makeWidgetDef(root, widget, parent));
        return Collections.unmodifiableList(ans);
    }
    
    /**
     * Retrieve the base widget types.
     * @return A list of the base widget types' bare classes.
     * @see Derived
     */
    public final List<Class<? extends WidgetDef>> getBaseWidgetDefClasses() {
        return getBaseWidgetDefClasses(this.getClass());
    }
    
    /**
     * Retrieve the base widget types for a given WidgetDef class.
     * @param cls The class in question.
     * @return A list of the base widget types' bare classes.
     * @see Derived
     */
    public final List<Class<? extends WidgetDef>> 
            getBaseWidgetDefClasses(Class<? extends WidgetDef> cls) {
        return getBaseWidgetDefClasses(cls, langDef);
    }
    
    /**
     * Retrieve the base widget types for a given WidgetDef class.
     * @param cls The class in question.
     * @param langDef The language definition for the class.
     * @return A list of the base widget types' bare classes.
     * @see Derived
     */
    public static List<Class<? extends WidgetDef>> 
            getBaseWidgetDefClasses(
                    Class<? extends WidgetDef> cls, LangDef langDef) {
        if (cls.equals(langDef.getRootWidgetType()))
            return Collections.emptyList();
        
        final List<Class<? extends WidgetDef>> bases, ans;
        
        final Derived annot = cls.getAnnotation(Derived.class);
        if (annot == null)
            bases = Collections.<Class<? extends WidgetDef>> emptyList();
        else
            bases = Arrays.asList(annot.value());
        
        final Class<? extends WidgetDef> root = langDef.getRootWidgetType();
        if (!bases.contains(root)) {
            ans = new ArrayList<Class<? extends WidgetDef>>(bases);
            ans.add(root);
        } else
            ans = bases;
        
        return Collections.unmodifiableList(ans);
    }
    
    private List<WidgetDef> ancestry = null;
    
    /** Get all widget types from which this one derives, in method resolution
     *  order, including this type itself. This order determines which types'
     *  members override members of which base types; in particular, a member
     *  of a type earlier in the list overrides that later in the list.
     *  (In the case of a single-inheritance type hierarchy, like Java's (not
     *  counting interfaces), this list will simply include this type, followed
     *  by its base type, and then that widget type's base, and so forth.)
     *  <p>
     *  The method resolution order (MRO) used here is called C3; I'm borrowing
     *  the algorithm from Python 2.3 (which in turn borrowed it from a Lisp
     *  variant called Dylan). See 
     *  {@link <a href="http://www.python.org/2.3/mro.html">this essay</a>} for
     *  a more thorough explanation and justification of the C3 algorithm.
     *  @return This type's ancestors, direct bases first, as WidgetDefs.
     *  @see <a href="http://www.python.org/2.3/mro.html">The Python 2.3 Method
     *  Resolution Order</a> */
    public final List<WidgetDef> getAncestry() {
        if (ancestry != null)
            return ancestry;
        
        if (this.getClass().equals(langDef.getRootWidgetType()))
            return Collections.singletonList(this);
        
        final List<WidgetDef> ans = new ArrayList<WidgetDef>();
        ans.add(this);
        
        final List<WidgetDef> bases = getBaseWidgetDefs();
        final List<List<WidgetDef>> baseAncestries =
            new ArrayList<List<WidgetDef>>(bases.size() + 1);
        
        for (WidgetDef base : bases)
            // We're going to need to modify this list, so make a copy
            baseAncestries.add(new LinkedList<WidgetDef>(base.getAncestry()));
        baseAncestries.add(new LinkedList<WidgetDef>(bases));
        
        // Perform what the Python essay refers to as the merge function
        
        MERGE: while (true) {
            boolean allEmpty = true;
            FIND_GOOD_HEAD: for (
                    List<WidgetDef> baseAncestry : baseAncestries) {
                if (baseAncestry.isEmpty())
                    continue;
                allEmpty = false;
                // Check to see if the head (first element) of baseAncestry
                // is a good head - i.e. it is not in the tail of any other
                // list in baseAncestries. (See the Python essay.)
                final WidgetDef head = baseAncestry.get(0);
                for (List<WidgetDef> otherAncestry : baseAncestries)
                    if (otherAncestry != baseAncestry
                            && otherAncestry.indexOf(head) > 0)
                        // This isn't a good head
                        continue FIND_GOOD_HEAD;
                // We made it through the whole loop; this is a good head
                // Accumulate it, then remove it from all of baseAncestries
                ans.add(head);
                for (List<WidgetDef> otherAncestry : baseAncestries)
                    otherAncestry.remove(head);
                continue MERGE;
            }
            if (allEmpty)
                break;
            else
                throw new BadWidgetDefinitionException(String.format(
                        "%s: inconsistent widget hierarchy", this
                        .getClass().getName()));
        }
        return (ancestry = ans);
    }
    
    /**
     * Test whether this widget definition derives from the given one.
     * @param widgetDef The potential ancestor of this widget definition.
     * @return {@code true} if this widget definition derives, directly or
     *      indirectly, from {@code widgetDef}.
     */
    public final boolean isDerivedFrom(WidgetDef widgetDef) {
        return getAncestry().contains(widgetDef);
    }
    
    /**
     * Test whether this widget definition derives from that specified by the
     * given WidgetDef class.
     * @param cls The class of the potential ancestor of this widget
     *      definition.
     * @return {@code true} if this widget definition derives, directly or
     *      indirectly, from that specified by {@code cls}.
     */
    public final boolean isDerivedFrom(Class<? extends WidgetDef> cls) {
        return getAncestry().contains(makeWidgetDef(cls, widget, parent));
    }
    
    /**
     * Determines whether one widget type is derived from or the same as
     * another, directly or otherwise.
     * @param deriv The type in question.
     * @param maybeBase The type which may be the base widget type.
     * @return {@code true} if {@code deriv} equals or derives from
     *         {@code maybeBase}.
     */
    protected final boolean isDerivedFrom(
            Class<? extends WidgetDef> deriv,
            Class<? extends WidgetDef> maybeBase) {
        return isDerivedFrom(deriv, maybeBase, langDef);
    }
    
    
    /**
     * Determines whether one widget type is derived from or the same as
     * another, directly or otherwise.
     * @param deriv The type in question.
     * @param maybeBase The type which may be the base widget type.
     * @param langDef The language definition in use.
     * @return {@code true} if {@code deriv} equals or derives from
     *         {@code maybeBase}.
     */
    protected static boolean isDerivedFrom(
            Class<? extends WidgetDef> deriv,
            Class<? extends WidgetDef> maybeBase,
            LangDef langDef) {
        if (deriv.equals(maybeBase))
            return true;
        List<Class<? extends WidgetDef>> bases = 
            getBaseWidgetDefClasses(deriv, langDef);
        if (bases.isEmpty())
            return false;
        if (bases.contains(maybeBase))
            return true;
        for (Class<? extends WidgetDef> base : bases)
            if (isDerivedFrom(base, maybeBase, langDef))
                return true;
        return false;
    }
    
    /**
     * Determines whether one widget type is derived from or the same as
     * another, directly or otherwise.
     * @param derivName The name of the type in question.
     * @param maybeBase The type which may be the base widget type.
     * @return {@code true} if {@code deriv} equals or derives from
     *         {@code maybeBase}.
     */
    protected final boolean isDerivedFrom(String derivName,
            Class<? extends WidgetDef> maybeBase) {
        return isDerivedFrom(findWidgetClass(derivName), maybeBase);
    }
    
    /**
     * Determines whether one widget type is derived from or the same as
     * another, directly or otherwise.
     * @param derivName The name of the type in question.
     * @param maybeBase The type which may be the base widget type.
     * @param langDef The language definition in use.
     * @return {@code true} if {@code deriv} equals or derives from
     *         {@code maybeBase}.
     */    
    protected static boolean isDerivedFrom(String derivName,
            Class<? extends WidgetDef> maybeBase, LangDef langDef) {
        return isDerivedFrom(findWidgetClass(derivName, langDef), 
                maybeBase, langDef);
    }

    /**
     * Determines whether one widget type is derived from or the same as
     * another, directly or otherwise.
     * @param derivName The name of the type in question.
     * @param maybeBaseName The name of the type which may be the base widget
     *        type.
     * @return {@code true} if {@code deriv} equals or derives from
     *         {@code maybeBase}.
     */
    protected final boolean isDerivedFrom(String derivName,
            String maybeBaseName) {
        return isDerivedFrom(findWidgetClass(derivName),
                findWidgetClass(maybeBaseName));
    }
    
    /**
     * Determines whether one widget type is derived from or the same as
     * another, directly or otherwise.
     * @param derivName The name of the type in question.
     * @param maybeBaseName The name of the type which may be the base widget
     *        type.
     * @param langDef The language definition in use.
     * @return {@code true} if {@code deriv} equals or derives from
     *         {@code maybeBase}.
     */
    protected static boolean isDerivedFrom(String derivName,
            String maybeBaseName, LangDef langDef) {
        return isDerivedFrom(findWidgetClass(derivName, langDef),
                findWidgetClass(maybeBaseName, langDef), langDef);
    }
    
    /** 
     * Annotation forbidding a widget type from being instantiated. Much like
     * an abstract class in Java, an abstract widget type can only be derived
     * from. However, an abstract widget type is more powerful in certain ways
     * than a concrete one, since all code-generating methods of an abstract
     * base widget type are inherited cumulatively, rather than just
     * getAttributes(). This means that an abstract widget type can declare
     * auxiliary variables and initialize them in the preamble, add code to
     * the postamble, etc. A concrete widget type cannot, for instance, add
     * variables in the getVariables() method, since it is responsible for
     * declaring the field with the id of the widget, which can only be done
     * once. 
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    @Documented
    public static @interface Abstract {}
    
    /**
     * Test whether this widget definition is for an abstract widget type.
     * @return {@code true} if this widget definition carries the
     *         {@link Abstract} annotation.
     * @see Abstract
     */
    public final boolean isAbstract() {
        return this.getClass().getAnnotation(Abstract.class) != null;
    }
    
    /**
     * Test whether a class defines an abstract widget type.
     * @param cls A subclass of {@link WidgetDef}.
     * @return {@code true} if the class carries the {@link Abstract}
     *         annotation.
     * @see Abstract
     */
    public static boolean isAbstract(Class<? extends WidgetDef> cls) {
        return cls.getAnnotation(Abstract.class) != null;
    }
    
    /**
     * Declare that a widget type cannot appear at the top level. For instance,
     * a layout widget without a parent in which to put the child widgets is
     * nonsensical. All widget types derived from this one will also be
     * contained (in other words, the {@lit @Contained} attribute is
     * inherited). 
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    @Documented
    public @interface Contained {}
    
    /**
     * Test whether the represented widget is contained, and so cannot appear
     * at the top level of a .gui file.
     * @return {@code true} if this widget definition (or any ancestor widget
     *         type) carries the {@link Contained} annotation.
     * @see Contained
     */
    public final boolean isContained() {
        return isContained(this.getClass());
    }
    
    /**
     * Test whether a class defines the type of a widget that is contained, and
     * so cannot appear at the top level of a .gui file.
     * @param cls A subclass of {@link WidgetDef}.
     * @return {@code true} if the class, or any ancestor widget type as
     *         declared by the {@link Derived} annotation, carries the
     *         {@link Contained} annotation.
     * @see Contained
     */
    public final boolean isContained(Class<? extends WidgetDef> cls) {
        return isContained(cls, langDef);
    }

    /**
     * Test whether a class defines the type of a widget that is contained, and
     * so cannot appear at the top level of a .gui file.
     * @param cls A subclass of {@link WidgetDef}.
     * @param langDef The language definition in use.
     * @return {@code true} if the class, or any ancestor widget type as
     *         declared by the {@link Derived} annotation, carries the
     *         {@link Contained} annotation.
     * @see Contained
     */
    public static boolean isContained(
            Class<? extends WidgetDef> cls, LangDef langDef) {
        // If the widget definition has the @Contained attribute, it's
        // certainly contained
        if (cls.getAnnotation(Contained.class) != null)
            return true;
        
        // Now search base classes
        for (Class<? extends WidgetDef> base : 
                getBaseWidgetDefClasses(cls, langDef))
            if (isContained(base, langDef))
                return true;
        
        return false;
    }
    
    /** Compare this WidgetDef to another.
     *  @param other The WidgetDef to compare to.
     *  @return {@code true} if this WidgetDef is of the same class as the
     *      given one and was given the same Widget at construction.
     */
    @Override
    public final boolean equals(final Object other) {
        if (!(other instanceof WidgetDef))
            return super.equals(other);
        return (this.getClass() == other.getClass() 
                && ((WidgetDef) other).widget == this.widget);
    }
    
    @Override
    public int hashCode() {
        return this.getClass().hashCode() ^ widget.hashCode();
    }

}
