package util;

import java.util.*;

/** Utility methods for specifying data concisely in code. The next-best thing
 *  to map and collection literals.
 *  <p>
 *  The purpose is easiest to see by example:
 *  <p><blockquote><pre>
 *      Map<Integer, String> numbers = Builders.buildMap(
 *          new TreeMap<Integer, String>())
 *          .put(0, "zero")
 *          .put(1, "one")
 *          .put(2, "two")
 *          .put(3, "three")
 *          .finishMap();
 *  </pre></blockquote> */
public final class Builders {
    /** An object with methods that add to a collection and can be chained
     *  together. See {@link Builders} for an example.
     *  @param <C> The class of the object being built.*/
    public interface Builder<C> {
        /** @return The built object. */
        C finish();
    }
    
    /** The return type of {@link Builders#buildMap}.
     *  @param <K> The type of the keys.
     *  @param <V> The type of the values.
     *  @param <C> The type of the map. 
     *  @see Builders#buildMap */
    public interface MapBuilder<K, V, C extends Map<?, ?>> 
            extends Builder<C> {
        /** @param key The key for the value.
         *  @param value The value to add to the map.
         *  @return <code>this</code>, so that calls may be chained. */
        MapBuilder<K, V, C> put(K key, V value);
        /** Add many entries with the same value.
         *  
         *  <p><b>NOTE:</b> The value must come <em>first</em>, since there are
         *  variably many key arguments and varargs must come last.
         *  
         *  @param value The value to add to the map.
         *  @param keys The keys for the value.
         *  @return <code>this</code>, so that calls may be chained. */
        MapBuilder<K, V, C> putForKeys(V value, K ... keys);
    }
    
    /** Build a map in one line of code.
     *  @param <K> The type of the keys.
     *  @param <V> The type of the values.
     *  @param <C> The type of the map.
     *  @param map The map to build onto.
     *  @return A {@link MapBuilder} whose put() method adds an entry to the
     *      map and returns itself so that calls can be chained together. */
    public static <K, V, C extends Map<? super K, ? super V>>
            MapBuilder<K, V, C> buildMap(final C map) {
        return new MapBuilder<K, V, C>() {
            public MapBuilder<K, V, C> put(K key, V value) {
                map.put(key, value);
                return this;
            }
            public MapBuilder<K, V, C> putForKeys(V value, K ... keys) {
                for (K key : keys)
                    map.put(key, value);
                return this;
            }
            public C finish() {
                return map;
            }
        };
    }
    
    /** The return type of {@link Builders#buildMultiMap}.
     *  @param <K> The type of the keys.
     *  @param <V> The type of the (individual) values.
     *  @param <C> The type of the map. 
     *  @see Builders#buildMultiMap */
    public interface MultiMapBuilder<K, V, C extends Map<?, ?>>
            extends MapBuilder<K, V, C> {
        /** @param key The key for the values.
         *  @param values Zero or more values; they will be put into a list
         *      and stored with the key in the map.
         *  @return <code>this</code>, so that calls may be chained. */
        MultiMapBuilder<K, V, C> put(K key, V ... values);
        
        // Redeclare a few things so they can return themselves as
        // MultiMapBuilders and not mere MapBuilders
        MultiMapBuilder<K, V, C> put(K key, V value);
        MultiMapBuilder<K, V, C> putForKeys(V value, K ... keys);
    }
    
    /** Build a map from things to immutable lists in one line of code.
     *  @param <K> The type of the keys.
     *  @param <V> The type of the (individual) values.
     *  @param <C> The type of the map.
     *  @param map The map to build onto.
     *  @return A {@link MapBuilder} whose put() method adds an entry to the
     *      map and returns itself so that calls can be chained together.
     *      All values in the map will be immutable lists. */
    public static <K, V, C extends Map<? super K, ? super List<V>>> 
            MultiMapBuilder<K, V, C> buildMultiMap(final C map) {
        return new MultiMapBuilder<K, V, C>() {
            public MultiMapBuilder<K, V, C> put(K key, V ... values) {
                map.put(key, Collections.unmodifiableList(
                        Arrays.asList(values)));
                return this;
            }
            public C finish() {
                return map;
            }
            public MultiMapBuilder<K, V, C> put(K key, V value) {
                map.put(key, Collections.singletonList(value));
                return this;
            }
            public MultiMapBuilder<K, V, C> putForKeys(V value, K... keys) {
                for (K key : keys)
                    put(key, value);
                return this;
            }
        };
    }
    
    /** The return type of {@link Builders#buildCollection}.
     *  @param <E> The type of the elements.
     *  @param <C> The type of the collection.
     *  @see Builders#buildCollection */
    public interface CollectionBuilder<E, C extends Collection<?>>
            extends Builder<C> {
        /** @param elems Zero or more items to add to the collection.
         *  @return <code>this</code>, so that calls may be chained. */
        CollectionBuilder<E, C> add(E ... elems);
    }
    
    /** Build a collection in one line of code.
     *  @param <E> The type of the elements.
     *  @param <C> The type of the collection.
     *  @param coll The collection to build onto.
     *  @return A {@link CollectionBuilder} whose add() method adds an
     *      element to the collection and returns itself so that calls can be
     *      chained together. */
    public static <E, C extends Collection<? super E>>
            CollectionBuilder<E, C> buildCollection(final C coll) {
        return new CollectionBuilder<E, C>() {
            public CollectionBuilder<E, C> add(E ... elems) {
                Collections.addAll(coll, elems);
                return this;
            }
            public C finish() {
                return coll;
            }
        };
    }
    
    /** Build a collection in one function call.
     *  @param <E> The type of the elements.
     *  @param <C> The type of the collection.
     *  @param coll The collection to build onto.
     *  @param elems Zero or more values to add to the collection. 
     *  @return {@code coll}, having had each element in {@code elems} added to
     *      it in order. */
    public static <E, C extends Collection<? super E>>
            C makeCollection(final C coll, E ... elems) {
        Collections.addAll(coll, elems);
        return coll;
    }
}
