package util;

import java.util.*;

/** A few utility methods for functional-ish programming, including concise
 *  string manipulation and composition. Along with {@link util.Builders},
 *  should be useful in cutting down on lengthy code duplication in library
 *  classes. */
public final class Functional {
    /** Concatenate the string represetations of several objects, with a
     *  separator.
     *  <p>
     *  See the example under {@link #join(Object[], String)}.
     *  @param <T> The type of the objects to concatenate.
     *  @param coll The objects to concatenate.
     *  @param sep The separator string.
     *  @return A string with the objects separated by the separator. */
    static public <T> String join(Collection<T> coll, String sep) {
        Iterator<T> iter = coll.iterator();
        if (!iter.hasNext())
            return "";
        StringBuilder builder = new StringBuilder(String.valueOf(iter.next()));
        while(iter.hasNext()) {
            builder.append(String.format("%s%s",
                    sep, String.valueOf(iter.next())));
        }
        return builder.toString();
    }
    
    /** Join an array into a string, separating the elements with the given
     *  separator string.
     *  <p><blockquote><pre>
     *      String[] things = { "eggs", "bacon", "spam" };
     *      System.out.println(Functional.join(things, ", ");
     *      System.out.println(Functional.join(things, " ... ");
     *  </pre></blockquote><p>
     *  Expected output:
     *  <p><blockquote><pre>
     *      eggs, bacon, spam
     *      eggs ... bacon ... spam
     *  </pre></blockquote>
     *  @param <T> The element type of the array.
     *  @param arr The array with the elements to concatenate.
     *  @param sep The string to separate the elements.
     *  @return The string concatenation of the elements of the array. */
    static public <T> String join(T[] arr, String sep) {
        return join(Arrays.asList(arr), sep);
    }
    
    /** Add a prefix and/or suffix to the string representation of each element
     *  in a collection.
     *  <p>
     *  See the example under {@link #wrapEach(Object[], String, String)}.
     *  @param <T> The element type of the collection.
     *  @param coll The collection.
     *  @param prefix The string to put before each element.
     *  @param suffix The string to put after each element.
     *  @return For each element, the prefix plus the element's string
     *      representation plus the suffix. */
    static public <T> String[] wrapEach (Collection<T> coll,
            String prefix, String suffix) {
        String[] ans = new String[coll.size()];
        for (Pair<Integer, T> entry : enumerate(coll))
            ans[entry.a] = String.format("%s%s%s", prefix, entry.b, suffix);
        return ans;
    }
    
    /** Add a prefix and/or suffix to the string representation of each element
     *  in any iterable.
     *  <p>
     *  See the example under {@link #wrapEach(Object[], String, String)}.
     *  @param <T> The element type of the iterable.
     *  @param coll The iterable.
     *  @param prefix The string to put before each element.
     *  @param suffix The string to put after each element.
     *  @return For each element, the prefix plus the element's string
     *      representation plus the suffix. */
    static public <T> String[] wrapEach (Iterable<T> coll,
            String prefix, String suffix) {
        ArrayList<String> ans = new ArrayList<String>();
        for (Pair<Integer, T> entry : enumerate(coll))
            ans.add(entry.a, String.format("%s%s%s", prefix, entry.b, suffix));
        final String[] empty = {}; 
        return ans.toArray(empty);
    }
    
    /** Add a prefix and/or suffix to the string representation of each element
     *  in an array.
     *  <p>
     *  For example, say we want to generate Java code to count from 1 to 10.
     *  <p><blockquote><pre>
     *      int[] numbers = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };
     *      String[] lines = Functional.wrapEach(numbers,
     *          "System.out.println(", ");");
     *      for (String line : lines)
     *          System.out.println(line);
     *  </pre></blockquote><p>
     *  Expected output (abbreviated):
     *  <p><blockquote><pre>
     *      System.out.println(1);
     *      System.out.println(2);
     *      ...
     *      System.out.println(10);
     *  </blockquote></pre>
     *  @param <T> The element type of the array.
     *  @param arr The array.
     *  @param prefix The string to put before each element.
     *  @param suffix The string to put after each element.
     *  @return For each element, the prefix plus the element's string
     *      representation plus the suffix. */
    static public <T> String[] wrapEach (T[] arr,
            String prefix, String suffix) {
        String[] ans = new String[arr.length];
        for (Pair<Integer, T> entry : enumerate(arr))
            ans[entry.a] = String.format("%s%s%s", prefix, entry.b, suffix);
        return ans;
    }
    
    /**
     * Put typographer's (Unicode) quotes around a collection of objects,
     * delimited by some separator.
     * @param <T> The element type of the collection.
     * @param coll The collection.
     * @param sep The string to separate the elements.
     * @return The joined, quoted, formatted string.
     */
    static public <T> String quoteAndJoin(Collection<T> coll, String sep) {
        return join(wrapEach(coll, "\u201c", "\u201d"), ", ");
    }
    
    /**
     * Put typographer's (Unicode) quotes around an array of objects, delimited
     * by some separator.
     * @param <T> The element type of the array.
     * @param coll The array.
     * @param sep The string to separate the elements.
     * @return The joined, quoted, formatted string.
     */
    static public <T> String quoteAndJoin(T[] coll, String sep) {
        return join(wrapEach(coll, "\u201c", "\u201d"), ", ");
    }
    
    /** An adaptor for iterating in a foreach loop, given an iterator.
     * 
     *  <p>(Really, Iterator&lt;T> should be a superinterface of Iterable&lt;T>,
     *  or at least iterators ought to be allowed in foreach loops (since the
     *  meaning is obvious). Oh well.) 
     *  @param <T> The type of object over which to iterate. */
    public static final class Iteration<T> implements Iterable<T> {
        /** The iterator iterated over. */
        public final Iterator<T> iter;
        
        /** Construct an Iterable for an Iterator.
         *  @param iter The Iterator to iterate over. */
        public Iteration(Iterator<T> iter) {
            this.iter = iter;
        }
        
        /** Get back the iterator passed in to the constructor.
         *  @return The iterator.
         *  @see java.lang.Iterable#iterator() */
        public Iterator<T> iterator() {
            return iter;
        }
        
        /** Convenience constructor (to avoid giving an explicit type).
         *  @param <T> The type of object over which to iterate.
         *  @param iter The iterator to iterate over.
         *  @return An iterable iterating using the iterator. */
        public static <T> Iteration<T> with(Iterator<T> iter) {
            return new Iteration<T>(iter);
        }
    }
    
    /** 
     * Iterate over a given iterator backwards.
     * <p>
     * Note that this is likely to be inefficient, as a list of the values
     * given by the iterator must be built at construction and then kept for
     * the duration of the iteration.
     * @param <T> The element type of the iterator.
     * @param iter The iterator to reverse.
     * @return An iterable which gives the same elements as {@code iter}, only
     *      backwards.
     */
    public static <T> Iterable<T> backward(Iterator<T> iter) {
        return backward(Iteration.with(iter));
    }
    
    /** 
     * Iterate over a given list backwards.
     * <p>
     * If the list is of a type that implements {@link java.util.RandomAccess},
     * this is efficient, since one need only go through the indices
     * backwards. Otherwise, a temporary ArrayList is required.
     * 
     * @param <T> The element type of the iterator.
     * @param list The list over which to iterate.
     * @return An iterable which gives the elements of {@code list} in reverse.
     */
    public static <T> Iterable<T> backward(final List<T> list) {
        // If it's a random-access list, this can be done efficiently by
        // counting through the indices backwards
        if (list instanceof RandomAccess)
            return new Iterable<T>() {
                public Iterator<T> iterator() {
                    return new Iterator<T>() {
                        int ix = list.size() - 1;
                        
                        public boolean hasNext() {
                            return ix >= 0;
                        }

                        public T next() {
                            return list.get(ix--);
                        }

                        public void remove() {
                            list.remove(ix + 1);
                        }
                    };
                }
            };
        
        // Otherwise, use an ArrayList as an intermediary
        return backward((Iterable<T>) list);
    }
    
    /**
     * Iterate over a given iterable backwards. Inefficient for large
     * iterables.
     * 
     * @param <T> The element type of the iterable.
     * @param coll The iterable to reverse.
     * @return An iterable which gives the elements that {@code coll} would
     *      give, only in reverse.
     */
    public static <T> Iterable<T> backward(final Iterable<T> coll) {
        final ArrayList<T> arrList = new ArrayList<T>();
        for (T elt : coll)
            arrList.add(elt);
        final Iterator<T> backIter = backward(arrList).iterator();
        return new Iterable<T>() {
            public Iterator<T> iterator() {
                return new Iterator<T>() {
                    private T lastReturned = null;
                    public boolean hasNext() {
                        return backIter.hasNext();
                    }
                    public T next() {
                        return (lastReturned = backIter.next());
                    }
                    public void remove() {
                        try {
                            ((Collection<T>) coll).remove(lastReturned);
                        } catch (ClassCastException e) {
                            throw new UnsupportedOperationException(e);
                        }
                    }
                };
            }
        };
    }
    
    /** An iterable representing a range of integers. Useful for iterating
     *  over integers in a foreach loop without using memory for the integers.
     *  (However, old-style ugly for loops are still more efficient since they
     *  can be done with only primitives.)
     *  <p>
     *  Has no public constructors; use {@link Functional#range(int, int)} or
     *  another such to obtain one. */
    public static final class Range implements Iterable<Integer> {
        /** @see Range */
        public final class RangeIterator implements Iterator<Integer> {
            private final int step;
            private final int stop;

            private int current;

            private RangeIterator(int step, int start, int stop) {
                super();
                this.step = step;
                this.stop = stop;
                current = start;
            }

            public boolean hasNext() {
                return current < stop;
            }

            public Integer next() {
                int temp = current;
                current += step;
                return temp;
            }

            public void remove() {
                throw new UnsupportedOperationException(
                        "attempted reordering of the integers");
            }
        }

        /** The beginning of the range. */
        public final int start;
        /** The first integer after the range. */
        public final int stop;
        /** The increment at each step. */
        public final int step;
        
        private Range(int start, int stop, int step) {
            this.start = start;
            this.stop = stop;
            this.step = step;
        }
        
        public RangeIterator iterator() {
            return new RangeIterator(step, start, stop);
        }
    }
    
    /** Construct the integer range from <code>start</code> (inclusive) to 
     *      <code>stop</code> (exclusive), in increments of <code>step</code>,
     *      as an efficient iterable.
     *  @param start The first integer in the range.
     *  @param stop The first integer outside of the range.
     *  @param step The increment between each element.
     *  @return An <code>Iterable&lt;Integer></code> over the range. */
    public static Range range(int start, int stop, int step) {
        return new Range(start, stop, step);
    }
    
    /**
     * Construct the integer range [<code>start</code>, <code>stop</code>),
     * as an efficient iterable.
     * @param start The first integer in the range.
     * @param stop The first integer outside the range.
     * @return An <code>Iterable&lt;Integer></code> over [<code>start</code>,
     *         <code>stop</code>).
     */
    public static Range range(int start, int stop) {
        return new Range(start, stop, 1);
    }
    
    /**
     * Construct the integer range [0, <code>stop</code>), as an efficient
     * iterable.
     * @param stop The first integer outside the range.
     * @return An {@code Iterable<Integer>} over [0, {@code stop}).
     */
    public static Range range(int stop) {
        return new Range(0, stop, 1);
    }
        
    /** An iterable over two iterables, pairing them element by element.
     *  @param <A> The first element type.
     *  @param <B> The second element type. */
    public static final class Zipper<A, B> implements Iterable<Pair<A, B>> {
        /** The first iterable, yielding the first element of each pair. */
        public final Iterable<A> collA;
        /** The second iterable, yielding the second element of each pair. */
        public final Iterable<B> collB;
        
        /** @param collA The iterable giving the first element of each pair.
         *  @param collB The iterable giving the second element of each pair.*/
        public Zipper(Iterable<A> collA, Iterable<B> collB) {
            this.collA = collA;
            this.collB = collB;
        }
        
        public Iterator<Pair<A, B>> iterator() {
            return new Iterator<Pair<A, B>>() {
                private Iterator<A> iterA = collA.iterator();
                private Iterator<B> iterB = collB.iterator();

                public boolean hasNext() {
                    return iterA.hasNext() && iterB.hasNext();
                }

                public Pair<A, B> next() {
                    return Pair.pair(iterA.next(), iterB.next());
                }

                public void remove() {
                    /* If one of them can be removed but not the other,
                     * that's fine. (Consider something returned by
                     * enumerate(), for example.) */
                    boolean success = false;
                    try {
                        iterA.remove();
                        success = true;
                    } catch (UnsupportedOperationException e) {}
                    try {
                        iterB.remove();
                    } catch (UnsupportedOperationException e) {
                        if (!success)
                            throw new UnsupportedOperationException();
                    }
                }
            };
        }        
    }
    
    /** Iterate over two iterables at once, pairing them element by element. 
     *  @param <A> The type of the first elements.
     *  @param <B> The type of the second elements.
     *  @param collA The first iterable over which to iterate.
     *  @param collB The second iterable.
     *  @return An iterable giving the elements from both, in pairs. */
    public static <A, B> Zipper<A, B> zip(
            Iterable<A> collA, Iterable<B> collB) {
        return new Zipper<A, B>(collA, collB);
    }
    
    /** Iterate over two iterators at once.
     *  @param <A> The element type of the first iterator.
     *  @param <B> The element type of the second iterator.
     *  @param iterA The first iterator.
     *  @param iterB The second iterator.
     *  @return An iterable giving the elements yielded by both iterators. */
    public static <A, B> Zipper<A, B> zip(
            Iterator<A> iterA, Iterator<B> iterB) {
        return new Zipper<A, B>(Iteration.with(iterA), Iteration.with(iterB));
    }
    
    /** Iterate over two arrays at once.
     *  @param <A> The element type of the first array.
     *  @param <B> The element type of the second array.
     *  @param arrA The first array.
     *  @param arrB The second array.
     *  @return An iterable giving the elements from both, in pairs. */
    public static <A, B> Zipper<A, B> zip(A[] arrA, B[] arrB) {
        return new Zipper<A, B>(Arrays.asList(arrA), Arrays.asList(arrB));
    }
    
    /** Iterate over (index, element) pairs corresponding to each element of
     *  an iterable with its index.
     *  @param <T> The element type.
     *  @param coll The collection to enumerate.
     *  @return <code>(index, element)</code> for <code>element ==
     *      coll.get(index)</code>. */
    public static <T> Iterable<Pair<Integer, T>> enumerate(Iterable<T> coll) {
        return zip(range(0, Integer.MAX_VALUE), coll);
    }

    /** Iterate over (index, element) pairs corresponding to each element of
     *  an array with its index.
     *  @param <T> The element type.
     *  @param arr The array to enumerate.
     *  @return <code>(index, element)</code> for <code>element ==
     *      arr[index]</code>. */
    public static <T> Iterable<Pair<Integer, T>> enumerate(T[] arr) {
        return zip(range(0, Integer.MAX_VALUE), Arrays.asList(arr));
    }

    /** Divide an array into those elements satisfying and those not
     *  satisfying a given predicate, returning two arrays.
     *  <p>
     *  For example, given the predicate <code>isEven</code> in the example
     *  from {@link Predicate}:
     *  <p><blockquote><pre>
     *      Integer[] numbers = { 1, 2, 3, 4, 5 };
     *      Pair&lt;Integer> evens_odds = Functional.separate(numbers, isEven);
     *      Integer[] evens = evens_odds.a, odds = evens_odds.b;
     *      System.out.format("Evens: %s\nOdds: %s\n", 
     *          Functional.join(evens, ", "), Functional.join(odds, ", "));
     *  </pre></blockquote><p>
     *  (see {@link #join(Object[], String)})
     *  <p>
     *  The output should be:
     *  <p><blockquote><pre>
     *      Evens: 2, 4
     *      Odds: 1, 3, 5
     *  </pre></blockquote>
     *
     *  @param <T> The element type of the array.
     *  @param arr The array to divide.
     *  @param pred The predicate to apply.
     *  @return A pair of arrays, containing respectively those elements
     *      satisfying and not satisfying the predicate. */
    @SuppressWarnings("unchecked")
    public static <T>
            Pair<T[], T[]> separate(T[] arr, Predicate<T> pred) {
        final ArrayList<T> tgtTrue = new ArrayList<T>(arr.length);
        final ArrayList<T> tgtFalse = new ArrayList<T>(arr.length);
        for (T item : arr)
            if (pred.f(item))
                tgtTrue.add(item);
            else
                tgtFalse.add(item);
        return Pair.of((T[]) tgtTrue.toArray(), (T[]) tgtFalse.toArray());
    }

    /** Divide a collection into those elements satisfying and those not
     *  satisfying a given predicate, returning two arrays.
     *  <p>
     *  See the example under 
     *  {@link #separate(Object[], Predicate)}.
     *
     *  @param <T> The element type of the collection.
     *  @param <C> The collection type.
     *  @param coll The collection to divide.
     *  @param pred The predicate to apply.
     *  @return A pair of arrays, containing respectively those elements
     *      satisfying and not satisfying the predicate. */
    @SuppressWarnings("unchecked")
    public static <T, C extends Collection<T>>
            Pair<T[], T[]> separate(C coll, Predicate<T> pred) {
        final ArrayList<T> tgtTrue = new ArrayList<T>(coll.size());
        final ArrayList<T> tgtFalse = new ArrayList<T>(coll.size());
        for (T item : coll)
            if (pred.f(item))
                tgtTrue.add(item);
            else
                tgtFalse.add(item);
        return Pair.of((T[]) tgtTrue.toArray(), (T[]) tgtFalse.toArray());
    }

    /** Divide an array into those elements satisfying and those not
     *  satisfying a given predicate.
     *  <p>
     *  For example, given the predicate <code>isEven</code> in the example
     *  from {@link Predicate}:
     *  <p><blockquote><pre>
     *      Integer[] numbers = { 1, 2, 3, 4, 5 };
     *      List&lt;Integer> evens, odds;
     *      evens = new ArrayList&lt;Integer>();
     *      odds = new ArrayList&lt;Integer>();
     *      Functional.separate(numbers, isEven, evens, odds);
     *      System.out.format("Evens: %s\nOdds: %s\n", evens, odds);
     *  </pre></blockquote><p>
     *  The output should be:
     *  <p><blockquote><pre>
     *      Evens: [2, 4]
     *      Odds: [1, 3, 5]
     *  </pre></blockquote>
     *
     *  @param <T> The element type of the array.
     *  @param <C1> The type of the first target collection.
     *  @param <C2> The type of the second target collection.
     *  @param arr The array to divide.
     *  @param pred The predicate to apply.
     *  @param tgtTrue The collection receiving those elements satisfying the
     *      predicate.
     *  @param tgtFalse The collection receiving those elements not satisfying
     *      the predicate.
     *  @return <code>Pair.of(tgtTrue, tgtFalse)</code>, as a convenience. */
    public static <T, C1 extends Collection<T>,
            C2 extends Collection<T>>
            Pair<C1, C2> separate(T[] arr, Predicate<T> pred,
                    C1 tgtTrue, C2 tgtFalse) {
        final Pair<C1, C2> tgt = Pair.of(tgtTrue, tgtFalse);
        for (T item : arr)
            if (pred.f(item))
                tgtTrue.add(item);
            else
                tgtFalse.add(item);
        return tgt;
    }

    /** Divide a collection into those elements satisfying and those not
     *  satisfying a given predicate.
     *  <p>
     *  See the example under 
     *  {@link #separate(Object[], Predicate, Collection, Collection)}.
     *
     *  @param <T> The element type of the collection.
     *  @param <C> The collection type.
     *  @param <C1> The type of the first target collection.
     *  @param <C2> The type of the second target collection.
     *  @param coll The collection to divide.
     *  @param pred The predicate to apply.
     *  @param tgtTrue The collection receiving those elements satisfying the
     *      predicate.
     *  @param tgtFalse The collection receiving those elements not satisfying
     *      the predicate.
     *  @return <code>Pair.of(tgtTrue, tgtFalse)</code>, as a convenience. */
    public static <T, C extends Collection<T>, C1 extends Collection<T>,
            C2 extends Collection<T>>
            Pair<C1, C2> separate(C coll, Predicate<T> pred,
                    C1 tgtTrue, C2 tgtFalse) {
        final Pair<C1, C2> tgt = Pair.of(tgtTrue, tgtFalse);
        for (T item : coll)
            if (pred.f(item))
                tgtTrue.add(item);
            else
                tgtFalse.add(item);
        return tgt;
    }
    
    /** Split an array at the first element not satisfying a predicate,
     *  returning two arrays.
     *  <p>
     *  For example, given the predicate <code>isEven</code> in the example
     *  from {@link Predicate}:
     *  <p><blockquote><pre>
     *      Integer[] numbers = { 2, 4, 6, 7, 8, 9, 10 };
     *      Pair&lt;Integer[], Integer[]> evenSpan_rest =
     *          Functional.span(numbers, isEven);
     *      Integer[] evenSpan = evenSpan_rest.a, rest = evenSpan.rest.b;
     *      System.out.format("Even span: %s\nRest: %s\n",
     *          Functional.join(evenSpan, ", "), Functional.join(rest, ", "));
     *  </pre></blockquote><p>
     *  (see {@link #join(Object[], String)})
     *  The output should be:
     *  <p><blockquote><pre>
     *      Even span: 2, 4, 6
     *      Rest: 7, 8, 9, 10
     *  </pre></blockquote>
     *  @param <T> The element type of the array.
     *  @param arr The array to split.
     *  @param pred The predicate to apply.
     *  @return <code>Pair.of(tgt1, tgt2)</code>, as a convenience. */
    static public <T>
            Pair<T[], T[]> span(T[] arr, Predicate<T> pred) {
        return span(Arrays.asList(arr), pred);
    }

    /** Split a collection at the first element not satisfying a predicate,
     *  returning two arrays.
     *  <p>
     *  See {@link #span(Object[], Predicate, Collection, Collection)} for an
     *  example.
     *  @param <T> The element type of the collection.
     *  @param <C> The collection type.
     *  @param coll The collection to split.
     *  @param pred The predicate to apply.
     *  @return <code>Pair.of(tgt1, tgt2)</code>, as a convenience. */
    @SuppressWarnings("unchecked")
    static public <T, C extends Collection<T>>
            Pair<T[], T[]> span(C coll, Predicate<T> pred) {
    
        Pair<ArrayList<T>, ArrayList<T>> lists =
            span(coll, pred, new ArrayList<T>(), new ArrayList<T>());
        return Pair.of((T[]) lists.a.toArray(), (T[]) lists.b.toArray());
    }

    /** Split an array at the first element not satisfying a predicate.
     *  <p>
     *  For example, given the predicate <code>isEven</code> in the example
     *  from {@link Predicate}:
     *  <p><blockquote><pre>
     *      List&lt;Integer> evenSpan, rest;
     *      Integer[] numbers = { 2, 4, 6, 7, 8, 9, 10 };
     *      evenSpan = new ArrayList&lt;Integer>();
     *      rest = new ArrayList&lt;Integer>();
     *      Functional.span(numbers, isEven, evenSpan, rest);
     *      System.out.format("Even span: %s\nRest: %s\n", evenSpan, rest);
     *  </pre></blockquote><p>
     *  The output should be:
     *  <p><blockquote><pre>
     *      Even span: [2, 4, 6]
     *      Rest: [7, 8, 9, 10]
     *  </pre></blockquote>
     *  @param <T> The element type of the array.
     *  @param <C1> The type of the first target collection.
     *  @param <C2> The type of the second target collection.
     *  @param arr The array to split.
     *  @param pred The predicate to apply.
     *  @param tgt1 The collection receiving the initial sublist of elements
     *      satisfying the predicate.
     *  @param tgt2 The collection receiving the rest.
     *  @return <code>Pair.of(tgt1, tgt2)</code>, as a convenience. */
    static public <T, C1 extends Collection<T>, C2 extends Collection<T>>
            Pair<C1, C2> span(T[] arr, Predicate<T> pred, C1 tgt1, C2 tgt2) {
        return span(Arrays.asList(arr), pred, tgt1, tgt2);
    }

    /** Split a collection at the first element not satisfying a predicate.
     *  <p>
     *  See {@link #span(Object[], Predicate, Collection, Collection)} for an
     *  example.
     *  @param <T> The element type of the collection.
     *  @param <C> The collection type.
     *  @param <C1> The type of the first target collection.
     *  @param <C2> The type of the second target collection.
     *  @param coll The collection to split.
     *  @param pred The predicate to apply.
     *  @param tgt1 The collection receiving the initial sublist of elements
     *      satisfying the predicate.
     *  @param tgt2 The collection receiving the rest.
     *  @return <code>Pair.of(tgt1, tgt2)</code>, as a convenience. */
    static public <T, C extends Collection<T>, C1 extends Collection<T>,
            C2 extends Collection<T>>
            Pair<C1, C2> span(C coll, Predicate<T> pred, C1 tgt1, C2 tgt2) {
        final Iterator<T> iter = coll.iterator();
        final Pair<C1, C2> tgt = Pair.of(tgt1, tgt2);
        if (!iter.hasNext())
            return tgt;
        
        T next = null;
        boolean exhausted = false;
        while (iter.hasNext()) {
            next = iter.next();
            if (!pred.f(next))
                break;
            tgt.a.add(next);
            if (!iter.hasNext())
                exhausted = true;
        }
        
        if (exhausted)
            return tgt;
        
        tgt.b.add(next);
        
        while (iter.hasNext()) {
            tgt.b.add(iter.next());
        }
        
        return tgt;
    }

    private Functional() {}
}
