package util;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Utility class which acts like a system shell manipulable from Java code.
 */
public class Shell {
    /** The current working directory. */
    protected File cwd = new File(System.getProperty("user.dir"));
    /** The stream to use for standard output. */
    protected PrintStream stdout = System.out;
    /** The stream to use for standard error. */
    protected PrintStream stderr = System.err;
    /** The stream to use for standard input. */
    protected InputStream stdin = System.in;
    /** The exit value of the last command executed. */
    protected int lastExitValue = 0;
    
    /**
     * Get the current directory. Analogous to {@code pwd}.
     * @return The current working directory.
     */
    public File getCwd() {
        return cwd;
    }
    
    /**
     * Change directory. Analogous to {@code cd}.
     * @param dir The directory to change to.
     * @return {@code this}, for chaining.
     * @throws IOException If {@code cwd} is not a directory or does not exist.
     */
    public synchronized Shell chdir(File dir) throws IOException {
        if (!dir.isDirectory())
            throw new IOException(dir.getPath() + ": no such directory");
        this.cwd = dir;
        return this;
    }

    /**
     * @return The stream to be used as standard error by processes spawned by
     * this shell.
     */
    public PrintStream getStdErr() {
        return stderr;
    }

    /**
     * @param stderr A stream which will receive anything a child process sends
     * to its standard error stream. By default this will be System.err, that
     * is, the system-wide standard error stream.
     * @return {@code this}, for chaining.
     */
    public synchronized Shell setStdErr(PrintStream stderr) {
        this.stderr = stderr;
        return this;
    }

    /**
     * @return The stream to be used as standard input by processes spawned by
     * this shell.
     */
    public InputStream getStdIn() {
        return stdin;
    }

    /**
     * @param stdin A stream to be connected to standard input for child
     * processes. By default this will be System.in, that is, the system-wide
     * standard input stream.
     * @return {@code this}, for chaining.
     */
    public synchronized Shell setStdIn(InputStream stdin) {
        this.stdin = stdin;
        return this;
    }

    /**
     * @return The stream to be used as standard output by processes spawned by
     * this shell.
     */
    public PrintStream getStdOut() {
        return stdout;
    }

    /**
     * @param stdout A stream which will receive anything a child process sends
     * to its standard output stream. By default this will be System.err, that
     * is, the system-wide standard output stream.
     * @return {@code this}, for chaining. 
     */
    public synchronized Shell setStdOut(PrintStream stdout) {
        this.stdout = stdout;
        return this;
    }
    
    /**
     * Run a process within this shell.
     * @param cmdLine The command line, split into its arguments. No shell-like
     *      processing - word splitting, variable substitution, etc. - will be
     *      performed.
     * @return The exit code returned by the program.
     * @throws IOException If the program cannot be found or run.
     * @throws InterruptedException If this thread is interrupted while the
     *      program is running.
     */
    public synchronized int run(String ... cmdLine) 
            throws IOException, InterruptedException {
        final Process process;
        try {
            process = 
                new ProcessBuilder(cmdLine)
                    .directory(cwd)
                    .start();
        } catch (IOException e) {
            stderr.println(e.getLocalizedMessage());
            return 1;
        }
        Streams.PipeHandler handler = 
            new Streams.PipeHandler(
                    stdin, process.getOutputStream(),
                    stdout, process.getInputStream(),
                    stderr, process.getErrorStream());
        handler.handle();
        final int exitValue = process.waitFor();
        handler.finish();
        return (lastExitValue = exitValue);
    }
    
    /**
     * Run a process within this shell, in the background.
     * @param cmdLine The command line, split into its arguments. No shell-like
     *      processing - word splitting, variable substitution, etc. - will be
     *      performed.
     * @return The process, or {@code null} if it could not be started.
     * @throws IOException If the program cannot be found or run.
     */
    public synchronized Process fork(final String ... cmdLine) 
            throws IOException {
        final Process process;
        try {
            process = 
                new ProcessBuilder(cmdLine)
                    .directory(cwd)
                    .start();
        } catch (IOException e) {
            stderr.println(e.getLocalizedMessage());
            return null;
        }
        final InputStream childOut = process.getInputStream();
        final InputStream childErr = process.getErrorStream();
        final OutputStream childIn = process.getOutputStream();
        final Streams.PipeHandler handler = 
            new Streams.PipeHandler(
                    stdin, childIn, 
                    stdout, childOut,
                    stderr, childErr);
        
        new Thread(Functional.join(cmdLine, " ")) {
            public void run() {
                try {
                    handler.handle();
                    try {
                        process.waitFor();
                    } catch (InterruptedException e) {}
                    handler.finish();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }.start();
        return process;
    }
    
    /**
     * Run a Java class as a process within this shell.
     * @param cls The class to run. Must declare a method with signature
     *      {@code public static void main(String[])}.
     * @param args Arguments to pass to the class's {@code main} method.
     * @return The exit value returned by the VM.
     * @throws IOException For any problems relating to I/O (not being able to
     *      access the class file, etc.)
     * @throws InterruptedException If this thread is interrupted while the
     *      class is running.
     * @throws IllegalArgumentException If {@code cls} does not declare a
     *      method with the correct signature.
     */
    public synchronized int runClass(Class<?> cls, String ... args)
            throws IOException, InterruptedException {
        boolean isInvalid = true;
        do {
            // Validate the class; it must have a method with signature
            // {@code public static void main(String[] args)}.
            try {
                Method main = cls.getMethod("main", String[].class);
                if (!Modifier.isStatic(main.getModifiers())
                        || !Modifier.isPublic(main.getModifiers())
                        || main.getReturnType() != void.class)
                    break;
            } catch (NoSuchMethodException e) {
                break;
            }
        } while (isInvalid = false);
        if (isInvalid) {
            throw new IllegalArgumentException(
                    String.format("%s has no suitable main() method",
                            cls.getName()));
        }
        String[] cmdLine = Builders.buildCollection(
                new ArrayList<String>(args.length + 4))
                        .add("java", "-cp",
                                System.getProperty("java.class.path"),
                                cls.getName())
                        .add(args)
                        .finish()
                        .toArray(new String[0]);
        return run(cmdLine);
    }
    
    /**
     * @return The exit value of the last process run by this shell.
     */
    public int getLastExitValue() {
        return lastExitValue;
    }
}