package util;

import java.io.*;

/**
 * Utility class for dealing with streams.
 */
public final class Streams {
    private static final byte buffer[] = new byte[4096];
    /**
     * Read everything from one stream into another.
     * @param in The stream to read from.
     * @param out The stream to write to.
     * @throws IOException If anything goes wrong on either end.
     */
    public static synchronized void pipeInto(
            InputStream in, OutputStream out) throws IOException {
        int numRead;
        while ((numRead = in.read(buffer)) != -1) {
            out.write(buffer, 0, numRead);
        }
    }
    
    /**
     * Controller for handling all three streams for a child process
     * asynchronously. Each of stdin, stdout, and stderr can be connected to
     * any suitable stream; the child process's corresponding streams will be
     * connected automatically in a multithreaded, asynchronous manner.
     */
    public static class PipeHandler {
        /**
         * The child's standard output stream, which, to the parent, is an
         * input stream.
         */
        public final BufferedInputStream childOut;
        /**
         * The child's standard error stream, which, to the parent, is an
         * input stream.
         */
        public final BufferedInputStream childErr;
        /**
         * An input stream supplied by the parent, to be connected to the
         * child's standard input.
         */
        public final BufferedInputStream parentIn;
        /**
         * The child's standard input stream, which, to the parent, is an
         * output stream.
         */
        public final BufferedOutputStream childIn;
        /**
         * An output stream supplied by the parent, to be connected to the
         * child's standard output.
         */
        public final BufferedOutputStream parentOut;
        /**
         * An output stream supplied by the parent, to be connected to the
         * child's standard error.
         */
        public final BufferedOutputStream parentErr;
        
        private volatile Thread inThread, outThread, errThread;
        /**
         * Construct a pipe handler.
         * @param parentIn The stream to use for input.
         * @param childIn The child's standard input stream.
         * @param parentOut The stream to use for output.
         * @param childOut The child's standard output stream.
         * @param parentErr The stream to use for stderr.
         * @param childErr The child's standard error stream.
         */
        public PipeHandler(InputStream parentIn, OutputStream childIn, 
                OutputStream parentOut, InputStream childOut,
                OutputStream parentErr, InputStream childErr) {
            this.childOut = 
                childOut == null ? null : new BufferedInputStream(childOut);
            this.parentIn = 
                parentIn == null ? null : new BufferedInputStream(parentIn);
            this.childIn = 
                childIn == null ? null : new BufferedOutputStream(childIn);
            this.childErr =
                childErr == null ? null : new BufferedInputStream(childErr);
            this.parentOut = 
                parentOut == null ? null : new BufferedOutputStream(parentOut);
            this.parentErr = 
                parentErr == null ? null : new BufferedOutputStream(parentErr);
        }
        
        /**
         * Run the pipe handler, forwarding data along the given connections.
         * @throws IOException If anything goes wrong with any of the streams.
         */
        public synchronized void handle() throws IOException {
            class PipeThread extends Thread {
                public final BufferedInputStream in;
                public final BufferedOutputStream out;
                
                PipeThread(BufferedInputStream in,
                        BufferedOutputStream out, String name) {
                    super(name);
                    this.in = in;
                    this.out = out;
                }

                public void run() {
                    if (in == null || out == null)
                        return;
                    try {
                        final byte[] buf = new byte[4092];
                        int numRead, avail;
                        try {
                            do { 
                                if ((avail = in.available()) == 0) {
                                    Thread.yield();
                                    continue;
                                }
                                numRead = in.read(buf, 0, avail);
                                out.write(buf, 0, numRead);
                            } while (!interrupted());
                        } finally {
                            if ((avail = in.available()) != 0) {
                                numRead = in.read(buf, 0, avail);
                                out.write(buf, 0, numRead);
                            }
                            out.flush();
                        }
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
            inThread = new PipeThread(parentIn, childIn, "Input Handler");
            outThread = new PipeThread(childOut, parentOut, "Output Handler");
            errThread = new PipeThread(childErr, parentErr, "Error Handler");

            inThread.start();
            outThread.start();
            errThread.start();
        }
        
        /**
         * Stop forwarding data, flushing all output.
         * @throws IOException If anything goes wrong.
         */
        public synchronized void finish() throws IOException {
            if (inThread != null) inThread.interrupt();
            if (outThread != null) outThread.interrupt();
            if (errThread != null) errThread.interrupt();
        }
    }
}