package util;

import java.io.*;

/**
 * A triple of objects of some specified types. Convenient for return values,
 * etc.
 * @param <A> The type of the first element of a triple.
 * @param <B> The type of the second element of a triple.
 * @param <C> The type of the third element of a triple.
 */
public class Triple<A, B, C> implements Serializable {
    /** The first element of the triple. */
    public final A a;
    /** The second element of the triple. */
    public final B b;
    /** The third element of the triple. */
    public final C c;
    
    /** @param a The first element of the triple.
     *  @param b The second element of the triple.
     *  @param c The third element of the triple.*/
    public Triple(A a, B b, C c) {
        this.a = a;
        this.b = b;
        this.c = c;
    }

    /** A convenience method so that the types of the elements don't have to
     *  be given explicitly.
     *  @param <A> The type of the first element. 
     *  @param <B> The type of the second element.
     *  @param <C> The type of the third element.
     *  @param a The first element of the triple.
     *  @param b The second element of the triple.
     *  @param c The third element of the triple.
     *  @return A newly constructed triple. */
    public static <A, B, C> Triple<A, B, C> triple(A a, B b, C c) {
        return new Triple<A, B, C>(a, b, c);
    }
    
    /** A more readable synonym for {@link #triple}.
     *  @param <A> The type of the first element. 
     *  @param <B> The type of the second element.
     *  @param <C> The type of the third element.
     *  @param a The first element of the triple.
     *  @param b The second element of the triple.
     *  @param c The third element of the triple.
     *  @return A newly constructed triple. */
    public static <A, B, C> Triple<A, B, C> of(A a, B b, C c) {
        return triple(a, b, c);
    }
    
    public boolean equals(Object other) {
        if (other instanceof Triple<?, ?, ?>) {
            final Triple<?, ?, ?> otherTriple = (Triple<?, ?, ?>) other;
            return a.equals(otherTriple.a) && 
                b.equals(otherTriple.b) &&
                c.equals(otherTriple.c);
        }
        return super.equals(other);
    }

    public int hashCode() {
        return Pair.rotL(a == null ? 0 : a.hashCode(), 8) ^
            Pair.rotL(b == null ? 0 : b.hashCode(), 16) ^
            Pair.rotL(c == null ? 0 : c.hashCode(), 24);
    }
    
    public String toString() {
        return String.format("(%s, %s, %s)",
                String.valueOf(a),
                String.valueOf(b),
                String.valueOf(c));
    }
}
